"""`validate` subcommand for `hopctl` (experimental)."""

from __future__ import annotations

import logging
import sys
import tempfile

from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeAlias

from rich import box
from rich.console import Group
from rich.live import Live
from rich.padding import Padding
from rich.rule import Rule
from rich.table import Table

import hoppr.main
import hoppr.models.validation
import hoppr.net
import hoppr.utils

from hoppr.cli.layout import HopprBasePanel, HopprSbomFilesLayout, console
from hoppr.logger import HopprLogger
from hoppr.models.sbom import Sbom
from hoppr.models.types import LocalFile, UrlFile
from hoppr.models.validation.base import BaseValidator
from hoppr.models.validation.code_climate import IssueList, IssueSeverity
from hoppr.models.validation.validators import validate

if TYPE_CHECKING:
    from rich.console import Console, ConsoleOptions, RenderResult
    from typing_extensions import TypedDict

    DictStrAny: TypeAlias = dict[str, Any]

    class IssuesDict(TypedDict):  # noqa: D101
        sbom: IssueList
        metadata: IssueList
        metadata_licenses: IssueList
        components: IssueList
        component_licenses: IssueList


all_issues: IssueList
layout: HopprSbomFilesLayout
logger: HopprLogger
summary_panel: HopprValidateSummary

_EXIT_CODE = 0


class HopprValidateSummary(HopprBasePanel):
    """`hopctl validate sbom` results summary."""

    summary_group = Group()

    def __init__(self, *args, title: str | None = "Summary", log_file: Path, **kwargs) -> None:
        self.log_file = log_file
        self.issues_found: dict[str, IssueList] = {}

        super().__init__(self.summary_group, title, *args, **kwargs)

    def __rich_console__(self, console_: Console, options: ConsoleOptions) -> RenderResult:
        issues_found = {sbom_file: issues for sbom_file, issues in self.issues_found.items() if issues}

        for sbom_file, issues in issues_found.items():
            issues_dict = _get_issues_dict(issues)

            self.summary_group.renderables.append(f"\n[bold]{sbom_file}\n")
            self._sbom_issues_table(issues_dict["sbom"])
            self._metadata_issues_table(issues_dict["metadata"])
            self._license_issues_table(issues_dict["metadata_licenses"])
            self._component_issues_table(issues_dict["components"])
            self._license_issues_table(issues_dict["component_licenses"])

            self.summary_group.renderables.append(Rule(style="[bold white]rule.line"))

        self.summary_group.renderables.append(f"See log file {self.log_file} for full results.")
        if hoppr.utils.is_basic_terminal():
            console_.line()
            console_.rule(title="Summary", characters="=", style="[bold]rule.line")
            console_.print(self.summary_group)
            return []

        return super().__rich_console__(console_, options)

    def _component_issues_table(self, issue_list: IssueList):
        if not issue_list:
            return

        table = Table(
            "Check",
            "Count",
            box=box.MINIMAL,
            title="Components",
            title_justify="left",
            title_style="bold",
        )
        issue_counts = defaultdict[str, int](int)

        for issue in issue_list:
            check_name = (issue.check_name or "").removeprefix("components.")
            issue_counts[check_name] += 1

        for check, count in sorted(issue_counts.items()):
            table.add_row(check, str(count))

        self.summary_group.renderables.append(Padding.indent(table, level=4))

    def _license_issues_table(self, issue_list: IssueList):
        if not issue_list:
            return

        table = Table(
            "Check",
            "Count",
            box=box.MINIMAL,
            title="Licenses",
            title_justify="left",
            title_style="bold",
        )

        issue_counts = defaultdict[str, int](int)

        for issue in issue_list:
            check_name = (issue.check_name or "").removeprefix("metadata.").removeprefix("components.")
            check_name = check_name.removeprefix("licenses.")
            issue_counts[check_name] += 1

        for check, count in sorted(issue_counts.items()):
            table.add_row(check, str(count))

        self.summary_group.renderables.append(Padding.indent(table, level=8))

    def _metadata_issues_table(self, issue_list: IssueList):
        if not issue_list:
            return

        table = Table(
            "Check",
            "Count",
            box=box.MINIMAL,
            title="Metadata",
            title_justify="left",
            title_style="bold",
        )

        for issue in issue_list:
            table.add_row((issue.check_name or "").removeprefix("metadata."), issue.description)

        self.summary_group.renderables.append(Padding.indent(table, level=4))

    def _sbom_issues_table(self, issue_list: IssueList):
        if not issue_list:
            return

        table = Table("Check", "Details", box=box.MINIMAL)

        for issue in issue_list:
            table.add_row((issue.check_name or "").removeprefix("sbom."), issue.description)

        self.summary_group.renderables.append(table)


def _init_globals(log_file: Path, verbose: bool):
    global all_issues, layout, logger, summary_panel

    all_issues = IssueList()
    logger = HopprLogger(
        name="hopctl-validate",
        filename=str(log_file),
        level=logging.DEBUG if verbose else logging.INFO,
    )

    layout = HopprSbomFilesLayout()

    summary_panel = HopprValidateSummary(log_file=log_file)


def _get_issues_dict(issues: IssueList) -> IssuesDict:
    issues_dict: IssuesDict = {
        "sbom": IssueList(),
        "metadata": IssueList(),
        "metadata_licenses": IssueList(),
        "components": IssueList(),
        "component_licenses": IssueList(),
    }

    for issue in issues:
        if not issue.check_name:
            continue

        match issue.check_name.split("."):
            case ["sbom", *_]:
                issues_dict["sbom"].append(issue)

            case ["metadata", "licenses", *_]:
                issues_dict["metadata_licenses"].append(issue)

            case ["metadata", *_]:
                issues_dict["metadata"].append(issue)

            case ["components", "licenses", *_]:
                issues_dict["component_licenses"].append(issue)

            case ["components", *_]:
                issues_dict["components"].append(issue)

    return issues_dict


def _get_result_status(issues: IssueList) -> str:
    result = "success"

    for issue in issues:
        if issue.severity == IssueSeverity.MINOR:
            result = "warn"
        elif issue.severity == IssueSeverity.MAJOR:
            global _EXIT_CODE
            _EXIT_CODE = 1

            result = "fail"
            break

    return result


def _start_live_display() -> Live:
    live_display = Live(layout, console=console, refresh_per_second=10)

    if not hoppr.utils.is_basic_terminal():
        live_display.start(refresh=True)

    return live_display


def _validate_sboms(sbom_files: list[Path], sbom_urls: list[str]):
    with tempfile.TemporaryDirectory() as tmpdir:
        for sbom_url in sbom_urls:
            temp_sbom_file = Path(tmpdir) / Path(sbom_url).name
            hoppr.net.download_file(url=sbom_url, dest=str(temp_sbom_file))
            sbom_files.append(temp_sbom_file)

        # Before loop to populate the SBOM files side panel
        for sbom_file in sbom_files:
            layout.add_job(description=sbom_file.name)

        BaseValidator.subscribe(observer=logger, callback=logger.log)

        for sbom_file in sbom_files:
            msg = f"Validating {sbom_file}..."
            logger.info(msg)
            layout.print(msg)

            file_result = "success"
            sbom_file_issues = validate(sbom_file)
            summary_panel.issues_found[str(sbom_file)] = sbom_file_issues
            all_issues.extend(sbom_file_issues)

            file_result = _get_result_status(sbom_file_issues)

            layout.update_job(name=Path(sbom_file).name, status=file_result)
            layout.stop_job(Path(sbom_file).name)
            layout.advance_job(Path(sbom_file).name)

        BaseValidator.unsubscribe(observer=logger)


def _validate_experimental(
    manifest_file: Path | None,
    sbom_files: list[Path],
    sbom_urls: list[str],
    output_file: Path | None,
    log_file: Path,
    verbose: bool,
):  # pragma: no cover
    """Validate SBOM file(s)."""
    _init_globals(log_file, verbose)

    if manifest_file:
        # Get all SBOMs that were included in the manifest file
        sbom_files.extend(file.local for file in Sbom.loaded_sboms if isinstance(file, LocalFile))
        sbom_urls.extend(file.url for file in Sbom.loaded_sboms if isinstance(file, UrlFile))

    live_display = _start_live_display()

    try:
        _validate_sboms(sbom_files, sbom_urls)
    finally:
        live_display.stop()

    console.print("\n", summary_panel)

    if output_file:
        output_file.write_text(
            data=all_issues.json(indent=2, by_alias=True, exclude_none=True, exclude_unset=True),
            encoding="utf-8",
        )

    sys.exit(_EXIT_CODE)



"""Collector plugin for Golang images."""

from __future__ import annotations

import base64
import hashlib
import itertools
import zipfile

from typing import TYPE_CHECKING

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx
import requests

from pydantic import SecretStr
from requests import HTTPError
from requests.auth import HTTPBasicAuth

import hoppr.net
import hoppr.utils

from hoppr import Hash, __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.exceptions import HopprPluginError
from hoppr.models.credentials import CredentialRequiredService, Credentials
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from os import PathLike
    from pathlib import Path

    from packageurl import PackageURL

    from hoppr.models.sbom import Component


class CollectGolangPlugin(
    SerialCollectorPlugin,
    supported_purl_types=["golang"],
    products=["golang/*"],
    system_repositories=["https://proxy.golang.org"],
):
    """Collector plugin for Golang images."""

    go_sum_url: str = "https://sum.golang.org/lookup"

    def get_version(self) -> str:  # noqa: D102
        return __version__

    def _download_package(
        self, purl: str | None, download_url: RepositoryUrl, dest_file: Path, creds: CredentialRequiredService | None
    ):
        """Downloads a golang package from a specific URL.

        Args:
            purl: Used to identify the package in the logs.
            download_url: Where the package will be downloaded from.
            dest_file: Where the package will be downloaded to.
            creds: Credentials used for the HTTP request.

        Raises:
            HopprPluginError: If package download fails.
        """
        response = hoppr.net.download_file(
            url=str(download_url),
            dest=str(dest_file),
            creds=creds,
            proxies=self._get_proxies_for(str(download_url)),
            timeout=self.process_timeout,
        )
        try:
            response.raise_for_status()
        except HTTPError as ex:
            msg = f"Failed to download golang package for {purl}, status_code={response.status_code}"
            self.get_logger().error(msg=msg, indent_level=2)
            raise HopprPluginError(msg) from ex

    def _get_package_checksum(self, metadata_url: str, purl: PackageURL, auth: HTTPBasicAuth | None) -> Hash:
        """Gets the package checksum from the metadata at the specified URL.

        Args:
            metadata_url: URL to the JSON document containing the package metadata.
            purl: PackageURL to get checksum for
            auth: Basic HTTP authentication

        Returns:
            The SHA256 hash contained in the metadata.

        Raises:
            HopprPluginError: If download fails.
            ValueError: If the downloaded file does not contain the checksum.
        """
        package_metadata_response = requests.get(
            url=metadata_url,
            auth=auth,
            timeout=300,
            proxies=self._get_proxies_for(metadata_url),
        )

        try:
            package_metadata_response.raise_for_status()
        except HTTPError as ex:
            msg = (
                f"Failed to download Golang package metadata from {metadata_url}, "
                f"status_code={package_metadata_response.status_code}"
            )
            raise HopprPluginError(msg) from ex

        response_text = package_metadata_response.text.split("\n")[1]

        if purl.name not in response_text and purl.version not in response_text:
            raise ValueError("Package metadata missing package information.")

        checksum = response_text.split("h1:")[1]

        decoded_checksum = base64.b64decode(checksum).hex()

        return Hash(alg=cdx.HashAlg.SHA_256, content=decoded_checksum)

    def _get_download_urls(self, purl: PackageURL, repo_url: str) -> tuple[str, RepositoryUrl, RepositoryUrl]:
        package_name = f"{'_'.join(filter(None, [purl.name, purl.version]))}.zip"
        golang_metadata_url = (
            RepositoryUrl(url=self.go_sum_url) / (purl.namespace or "") / f"{purl.name}@{purl.version}"
        )
        golang_download_url = (
            RepositoryUrl(url=repo_url) / (purl.namespace or "") / purl.name / "@v" / f"{purl.version}.zip"
        )

        return package_name, golang_metadata_url, golang_download_url

    def _get_golang_file_hash(self, artifact: str | PathLike[str]) -> Hash:
        """Compute hash of downloaded golang component using algorithm used for golang package checksum database.

        Golang documentation explaining algorithm:
            - https://pkg.go.dev/golang.org/x/mod/sumdb/dirhash#Hash1

        Golang algorithm in code:
            - https://cs.opensource.google/go/x/mod/+/v0.16.0:sumdb/dirhash/hash.go;l=44.

        Args:
            artifact: Path to downloaded file

        Returns:
            A Hash Object with computed hexidecimal hash digest.

        Raises:
            HopprPluginError: If artifact is not a Zip file.
        """
        if not zipfile.is_zipfile(artifact):
            raise HopprPluginError("Incorrect File Type for golang hash generation")

        with zipfile.ZipFile(artifact) as pkg_zip:
            file_list = pkg_zip.namelist()
            file_list.sort()
            hash_obj = hashlib.sha256()

            for file in file_list:
                if "\n" in file:
                    raise HopprPluginError("Filenames with newlines are not supported")

                file_content = pkg_zip.read(file)
                file_hash = hashlib.sha256()
                file_hash.update(file_content)

                file_entry = f"{file_hash.hexdigest()}  {file}\n"
                hash_obj.update(str.encode(file_entry))

        return Hash(alg=cdx.HashAlg.SHA_256, content=hash_obj.hexdigest())

    def _update_hashes(self, comp: Component, package_path: Path) -> Result:
        """Update the hashes of a component with hashes generated from its package file.

        Args:
            comp: Component object whose hashes will be updated.
            package_path: Path of the downloaded package file for the provided component.

        Raises:
            HopprPluginError: If an existing hash doesn't match.
        """
        generated_hashes = self._get_artifact_hashes(package_path)
        golang_hash = self._get_golang_file_hash(package_path)
        generated_hashes = itertools.chain(generated_hashes, [golang_hash])

        try:
            comp.update_hashes(generated_hashes)
        except ValueError as ex:
            raise HopprPluginError(f"Hash for {comp.name} does not match expected hash.") from ex

        return Result.success()

    def _validate_hashes(
        self, comp: Component, metadata_url: RepositoryUrl, purl: PackageURL, auth: HTTPBasicAuth | None
    ) -> Result:
        """Validates the component hashes against the hashes in the package metadata provided by the golang server.

        Args:
            comp: Component object whose hashes will be validated.
            metadata_url: URL to the JSON document containing the package metadata.
            purl: PackageURL to validate hashes for
            auth: Basic HTTP authentication

        Raises:
            HopprPluginError: If hash validation fails.
        """
        try:
            validation_hash = self._get_package_checksum(metadata_url.url, purl, auth=auth)
        except (HopprPluginError, ValueError) as ex:
            self.get_logger().error(msg=str(ex), indent_level=2)
            raise HopprPluginError(f"Failed to download Golang package metadata from {metadata_url}.") from ex

        if not comp.contains_hashes([validation_hash]):
            raise HopprPluginError(f"Hash validation failed for {comp.name}.")

        return Result.success()

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """Download golang package via get request."""
        purl = hoppr.utils.get_package_url(comp.purl)

        package_name, golang_metadata_url, golang_download_url = self._get_download_urls(purl, repo_url)
        self.get_logger().info(msg=f"Copying Golang package from {comp.purl}", indent_level=2)

        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        package_out_path = target_dir / package_name

        auth: HTTPBasicAuth | None = None
        if (creds := Credentials.find(f"{repo_url}")) and isinstance(creds.password, SecretStr):
            auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())

        try:
            self._download_package(comp.purl, golang_download_url, package_out_path, creds)
            self._update_hashes(comp, package_out_path)
            self._validate_hashes(comp, golang_metadata_url, purl, auth)
        except HopprPluginError as ex:
            return Result.fail(message=str(ex))

        self.set_collection_params(comp, repo_url, package_out_path)
        return Result.success(return_obj=comp)



"""Exclude models for `hopctl validate sbom` configuration file."""

from __future__ import annotations

import re

from typing import TYPE_CHECKING, TypeGuard

import jmespath.functions

from pydantic import root_validator

from hoppr.models.validation.base import BaseExcludeConfig

if TYPE_CHECKING:
    from rich.repr import RichReprResult


class JMESPathOptions(jmespath.Options):
    """Custom JMESPath options and function to enable evaluation of regular expressions."""

    class JMESPathRegexFunction(jmespath.functions.Functions):
        """Custom JMESPath function to enable evaluation of regular expressions."""

        @jmespath.functions.signature({"types": ["string", "null"]}, {"types": ["string"]})
        def _func_regexp(self, string: str, pattern: str) -> bool:
            """Enables use of a custom `regexp` function inside a JMESPath search expression.

            NOTE: the `_func_` method name prefix is required.

            Args:
                string: The string to search
                pattern: The pattern to match against

            Returns:
                `True` if a match was found, otherwise `False`
            """
            if not string or not (
                pattern_match := re.match(
                    pattern="(?:regexp:)?/(?P<pattern>.*)/(?P<flags>[GMIgmi]+)?",
                    string=pattern,
                )
            ):
                return False

            pattern = pattern_match["pattern"]

            re_func = re.match
            re_flags = re.RegexFlag(0)

            for flag in pattern_match["flags"] or "":
                if flag.upper() == "G":
                    re_func = re.search
                else:
                    re_flags |= re.RegexFlag[flag.upper()]

            return re_func(pattern=pattern, string=string, flags=re_flags) is not None

    def __init__(self):
        super().__init__(custom_functions=self.JMESPathRegexFunction())


class ExcludeConfig(BaseExcludeConfig):
    """Configuration for items to exclude from validation."""

    def __repr__(self) -> str:  # pragma: no cover
        attributes = ", ".join([f"{key}={value}" for key, value in self.__dict__.items() if value])
        return f"ExcludeConfig({attributes})"

    def __rich_repr__(self) -> RichReprResult:  # pragma: no cover
        yield from [(key, value) for key, value in self.__dict__.items() if value]

    @classmethod
    def _build_jmespath(cls, obj: object) -> list[str]:
        expr: list[str] = []

        match obj:
            case dict():
                expr.extend(cls._build_jmespath(list(obj.items())))

            case list():
                expr.extend(token for value in obj for token in cls._build_jmespath(value))

            case (str() as key, dict() as value):
                expr.extend(
                    token
                    for subkey, subvalue in value.items()
                    for token in cls._build_jmespath((f'"{key}"."{subkey}"', subvalue))
                )

            case (str() as key, list() as values):
                expr.extend([f'"{key}"', "[?", *cls._build_jmespath(values), "]"])

            case (str() as key, str() as value) if value.startswith("regexp:"):
                expr.append(f"""regexp("{key}", '{value}')""")

            case (*_, str() as value) if value.startswith("jmespath:"):
                expr.append(value.removeprefix("jmespath:"))

            case (str() as key, str() as value):
                expr.extend([key, "==", f"'{value}'"])

        return expr

    @root_validator
    @classmethod
    def convert_to_jmespath(cls, values: dict[str, list[str | object] | None]) -> dict[str, list[str]]:
        """Convert complex objects to JMESPath expressions."""

        def _has_items(obj: tuple[str, list[object] | None]) -> TypeGuard[tuple[str, list[object]]]:
            return bool(obj[1])

        converted: dict[str, list[str]] = {
            field_name: [
                item if isinstance(item, str) and item.startswith("jmespath:")
                else f"jmespath:[?{''.join(cls._build_jmespath(item))}]"
                for item in items
            ]
            for field_name, items in filter(_has_items, values.items())
        }  # fmt: skip

        return converted



"""Collector plugin for Gem packages."""

from __future__ import annotations

from typing import TYPE_CHECKING

import jmespath
import requests as requests

from requests import HTTPError
from typing_extensions import override

import hoppr.net
import hoppr.utils

from hoppr import Hash, __version__, cdx
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.exceptions import HopprPluginError
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from pathlib import Path

    from hoppr.models.credentials import CredentialRequiredService
    from hoppr.models.sbom import Component


class CollectGemPlugin(
    SerialCollectorPlugin,
    supported_purl_types=["gem"],
    products=["gem/*"],
    system_repositories=["https://rubygems.org/api/v2/rubygems"],
):
    """Collector plugin for Gem packages."""

    def get_version(self) -> str:  # noqa: D102
        return __version__

    def _update_hashes(self, comp: Component, package_path: Path) -> Result:
        generated_hashes = self._get_artifact_hashes(package_path)
        try:
            comp.update_hashes(generated_hashes)
        except ValueError as ex:
            raise HopprPluginError(f"Hash for {comp.name} does not match expected hash.") from ex

        return Result.success()

    def _validate_hashes(self, comp: Component, checksum: Hash) -> Result:
        if not comp.contains_hashes([checksum]):
            raise HopprPluginError(f"Hash validation failed for {comp.name}.")

        return Result.success()

    def _get_package_metadata(self, api_url: RepositoryUrl) -> requests.Response:
        response = requests.get(str(api_url), timeout=60)

        try:
            response.raise_for_status()
        except HTTPError as ex:
            msg = f"RubyGems API web request error, request URL='{api_url}', status_code={response.status_code}"
            self.get_logger().error(msg=msg, indent_level=2)
            raise HopprPluginError(msg) from ex

        return response

    def _download_gem_file(self, download_url: str, package_out_path: Path, comp: Component):
        response = hoppr.net.download_file(url=download_url, dest=str(package_out_path), timeout=self.process_timeout)

        if not Result.from_http_response(response=response, return_obj=comp).is_success():
            msg = f"Failed to download Gem package {comp.purl}, status_code={response.status_code}"
            self.get_logger().error(msg=msg, indent_level=2)
            raise HopprPluginError(msg)

    @override
    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        purl = hoppr.utils.get_package_url(comp.purl)

        api_url = RepositoryUrl(url=repo_url) / purl.name / "versions" / f"{purl.version}.json"

        if platform := purl.qualifiers.get("platform"):
            api_url = RepositoryUrl(url=f"{api_url}?platform={platform}")

        try:
            gem_package_metadata = self._get_package_metadata(api_url)
        except HopprPluginError as ex:
            return Result.fail(message=str(ex))

        checksum = Hash(
            alg=cdx.HashAlg.SHA_256, content=jmespath.search(expression="sha", data=gem_package_metadata.json())
        )

        download_url: str | None = jmespath.search(expression="gem_uri", data=gem_package_metadata.json())

        if not download_url:
            msg = f"Unable to retrieve download URL for Gem '{comp.purl}'"
            self.get_logger().error(msg=msg, indent_level=2)
            return Result.fail(message=msg)

        self.get_logger().info(msg=f"Copying Gem package from {api_url}", indent_level=2)

        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        self.get_logger().info(msg=target_dir.as_posix())

        package_out_path = target_dir / f"{purl.name}-{purl.version}.gem"

        self.get_logger().info(msg="Downloading Gem:", indent_level=2)
        self.get_logger().info(msg=f"source: {download_url}", indent_level=3)
        self.get_logger().info(msg=f"destination: {package_out_path}", indent_level=3)

        try:
            self._download_gem_file(download_url, package_out_path, comp)
            self._update_hashes(comp, package_out_path)
            self._validate_hashes(comp, checksum)
        except HopprPluginError as ex:
            return Result.fail(message=str(ex))

        self.set_collection_params(comp, repo_url, package_out_path)
        return Result.success(return_obj=comp)



"""Collector plugin for RPM packages."""

from __future__ import annotations

import gzip
import os
import re
import time
import warnings

from collections import OrderedDict
from configparser import ConfigParser
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Final

import jmespath
import requests
import xmltodict

from pydantic import AnyUrl, Extra, Field, FileUrl, HttpUrl, SecretStr, root_validator, validator
from requests import HTTPError, Response
from requests.auth import HTTPBasicAuth
from typing_extensions import Self, override

import hoppr.net
import hoppr.utils

from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_process, hoppr_rerunner
from hoppr.constants import BomProps
from hoppr.exceptions import HopprExperimentalWarning, HopprPluginError
from hoppr.models import HopprContext
from hoppr.models.base import ExtendedProto, HopprBaseModel
from hoppr.models.credentials import CredentialRequiredService, Credentials
from hoppr.models.manifest import SearchSequence
from hoppr.models.sbom import Component, Hash
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable, MutableMapping

    from packageurl import PackageURL
    from pydantic.typing import DictStrAny
    from rich.repr import RichReprResult


warnings.filterwarnings(action="once", category=HopprExperimentalWarning)
warnings.warn(
    message="The RPM collector plugin is experimental; use at your own risk.",
    category=HopprExperimentalWarning,
    stacklevel=2,
)


def _file_exists(path: Path) -> Path:
    if not path.is_file():
        raise ValueError(f"File does not exist: {path}")

    return path


class RepoConfig(HopprBaseModel, extra=Extra.allow):
    """Configuration options for CollectRpmPlugin."""

    base_url: list[str] = Field(alias="baseurl")
    mirror_list: HttpUrl | FileUrl | AnyUrl | str | None = Field(default=None, alias="mirrorlist")
    enabled: bool = True
    gpg_check: bool = Field(default=True, alias="gpgcheck")
    gpg_key: str | None = Field(default=None, alias="gpgkey")
    name: str | None = None
    priority: int = 1
    proxy: str = "_none_"
    username: str | None = None
    password: str | None = None
    ssl_ca_cert: Path | None = Field(default=None, alias="sslcacert")
    ssl_client_cert: Path | None = Field(default=None, alias="sslclientcert")
    ssl_client_key: Path | None = Field(default=None, alias="sslclientkey")

    vars_dir: ClassVar[Path | None] = None
    base_arch: ClassVar[str | None] = None
    release_ver: ClassVar[str | None] = None

    _validate_ssl_ca_cert = validator("ssl_ca_cert", allow_reuse=True)(_file_exists)
    _validate_ssl_client_cert = validator("ssl_client_cert", allow_reuse=True)(_file_exists)
    _validate_ssl_client_key = validator("ssl_client_key", allow_reuse=True)(_file_exists)

    @validator("base_url", pre=True)
    def _assemble_base_url(cls, base_url: str | list[str]) -> list[str]:
        return [base_url] if isinstance(base_url, str) and not isinstance(base_url, list) else base_url

    @validator("enabled", pre=True)
    def _assemble_enabled(cls, enabled: str | bool | int) -> bool:
        return False if str(enabled).lower() == "false" else bool(enabled)

    @root_validator(allow_reuse=True)
    def _resolve_dnf_vars(cls, values: DictStrAny) -> DictStrAny:
        if cls.vars_dir is not None and not cls.vars_dir.exists():
            return values

        for key, value in values.items():
            if isinstance(value, str):
                value = cls._match_dnf_var(value)
                values[key] = value
            elif isinstance(value, list):
                for index, item in enumerate(value):
                    value[index] = cls._match_dnf_var(item)
                values[key] = value

        return values

    @classmethod
    def _match_dnf_var(cls, content: str) -> str:
        matches = re.findall(r"\$\w+", content)

        for var in matches:
            if var == "$basearch" and cls.base_arch:
                content = content.replace(var, cls.base_arch)
            elif var == "$releasever" and cls.release_ver:
                content = content.replace(var, cls.release_ver)
            elif cls.vars_dir:
                content = content.replace(var, (cls.vars_dir / var.removeprefix("$")).read_text())

        return content


class RepoConfigList(HopprBaseModel):
    """List of RepoConfig objects."""

    __root__: list[RepoConfig]

    def __getitem__(self, idx: int) -> RepoConfig:
        return self.__root__[idx]  # pragma: no cover

    def __init__(self, issues: Iterable[RepoConfig] | None = None, **data):
        data["__root__"] = data.get("__root__", list(issues or []))
        super().__init__(**data)

    def __iter__(self):
        yield from self.__root__

    def __len__(self) -> int:
        return len(self.__root__)  # pragma: no cover

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__root__})"  # pragma: no cover

    def __rich_repr__(self) -> RichReprResult:
        yield self.__root__  # pragma: no cover

    def append(self, item: RepoConfig) -> None:
        """Append object to the end of the list."""
        self.__root__.append(item)  # pragma: no cover

    def extend(self, iterable: Iterable[RepoConfig]) -> None:
        """Extend list by appending elements from the iterable."""
        self.__root__.extend(iterable)  # pragma: no cover

    @classmethod
    def parse_file(
        cls,
        path: str | Path,
        *,
        content_type: str | None = None,
        encoding: str = "utf-8",
        proto: ExtendedProto | None = None,
        allow_pickle: bool = False,
    ) -> Self:
        """Parse a .repo file."""
        parser = ConfigParser()
        parser.read(path)

        repo_config_list = cls()

        # Use ConfigParser to parse values in file into new object
        for section in parser.sections():
            if bool(parser[section].get("enabled", None)):
                repo_config_list.append(RepoConfig.parse_obj(dict(parser[section])))

        return repo_config_list


class CollectRpmConfig(HopprBaseModel):
    """Configuration options for CollectRpmPlugin."""

    base_arch: str | None = Field(default=None, description="Value for $basearch in repository configs")
    release_ver: str | None = Field(default=None, description="Value for $releasever in repository configs")
    vars_dir: Path | None = Field(
        default=next((Path("/", "etc", name, "vars") for name in ["dnf", "yum"]), None),
        description="DNF variable definitions directory.",
    )


class CollectRpmPlugin(
    SerialCollectorPlugin,
    products=["rpm/*"],
    supported_purl_types=["rpm"],
    system_repositories=[],
):
    """Collector plugin for RPM packages."""

    REQUEST_RETRIES: Final[int] = 3
    REQUEST_RETRY_INTERVAL: Final[float] = 5
    REQUEST_TIMEOUT: Final[float] = 60

    rpm_data: ClassVar[MutableMapping[str, OrderedDict[str, Any]]] = {}

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.collector_config = CollectRpmConfig(**(config or {}))
        self.repo_configs = RepoConfigList()

    def get_version(self) -> str:
        """Get plugin version."""
        return hoppr.__version__

    def _update_hashes(self, comp: Component, package_path: Path):
        """Update the hashes of a component with hashes generated from its package file.

        Args:
            comp: Component object whose hashes will be updated.
            package_path: Path of the downloaded package file for the provided component.

        Raises:
            HopprPluginError: If an existing hash doesn't match.
        """
        generated_hashes = self._get_artifact_hashes(package_path)
        try:
            comp.update_hashes(generated_hashes)
        except ValueError as ex:
            raise HopprPluginError(f"Hash for {comp.name} does not match expected hash.") from ex

    def _download_component(
        self,
        download_url: str,
        dest_file: Path,
        creds: CredentialRequiredService | None = None,
        cert: tuple[str, str] | None = None,
    ):
        self.get_logger().info(msg="Downloading RPM package:", indent_level=2)
        self.get_logger().info(msg=f"source: {download_url}", indent_level=3)
        self.get_logger().info(msg=f"destination: {dest_file}", indent_level=3)

        # Download the RPM file to the target directory
        response = hoppr.net.download_file(
            url=download_url,
            dest=str(dest_file),
            creds=creds,
            proxies=self._get_proxies_for(download_url),
            cert=cert,
        )

        if not (download_result := Result.from_http_response(response=response)).is_success():
            raise HopprPluginError(download_result.message)

    def _determine_sha_hash_alg(self, hash_alg: str, hash_string: str) -> hoppr.net.HashlibAlgs:
        """Determine type of sha hash if not specified."""
        match len(hash_string):
            case 32:
                return "md5"
            case 40:
                return "sha1"
            case 56:
                return "sha224"
            case 64:
                return "sha256"
            case 96:
                return "sha384"
            case 128:
                return "sha512"
            case _:
                raise HopprPluginError(f'Could not determine hash algorithm for hash type "{hash_alg}"')

    def _get_download_url(self, purl: PackageURL, repo_url: str) -> tuple[str, hoppr.net.HashlibAlgs, str]:
        """Get information required to download and verify an RPM package.

        Args:
            purl: PackageURL of component attributes
            repo_url: The RPM repository URL

        Returns:
            Tuple containing download URL, hash algorithm, and checksum string.

        Raises:
            HopprPluginError: malformed/missing `purl.version`, or component not found in repodata
        """
        try:
            rpm_version, rpm_release = purl.version.split("-")
        except (AttributeError, ValueError) as ex:
            raise HopprPluginError(f"Failed to parse version string from PURL: '{purl}'") from ex

        arch = purl.qualifiers.get("arch") or "noarch"

        if (check_result := self.check_purl_specified_url(purl, repo_url)).is_fail():
            if self.context.strict_repos:
                raise HopprPluginError(check_result.message)

            repo_url = purl.qualifiers.get("repository_url", repo_url)
            self._populate_rpm_data(RepositoryUrl(url=repo_url))

        self.get_logger().debug("Searching RPM data for package with attributes:", indent_level=2)
        self.get_logger().debug("name:    %s", purl.name, indent_level=3)
        self.get_logger().debug("version: %s", rpm_version, indent_level=3)
        self.get_logger().debug("release: %s", rpm_release, indent_level=3)
        self.get_logger().debug("arch:    %s", arch, indent_level=3)

        component_data = jmespath.search(
            expression=f"""
                metadata.package[? name=='{purl.name}' &&
                version."@ver"=='{rpm_version}' &&
                version."@rel"=='{rpm_release}' &&
                (arch=='{arch}' || arch=='noarch')] | [0]
            """,
            data=self.rpm_data[repo_url],
        )

        if not component_data:
            raise HopprPluginError(f"RPM package not found in repository: '{purl}'")

        download_url, hash_alg, hash_string = jmespath.search(
            expression='[location."@href", checksum."@type", checksum."#text"]',
            data=component_data,
        )

        download_url = f"{repo_url}/{download_url}"

        if hash_alg not in hoppr.net.HASH_ALG_MAP.values():
            self._logger.info('Unknown hash algorithm "%s". Attempting to resolve', hash_alg, indent_level=1)
            try:
                hash_alg = self._determine_sha_hash_alg(hash_alg, hash_string)
                self._logger.info('Resolved hash algorithm to "%s"', hash_alg, indent_level=1)
            except HopprPluginError as e:
                raise HopprPluginError(e) from e

        return download_url, hash_alg, hash_string

    def _get_primary_xml_data(
        self,
        repo_url: RepositoryUrl,
        repomd_dict: OrderedDict[str, Any],
        auth: HTTPBasicAuth | None = None,
        cert: tuple[str, str] | None = None,
    ) -> OrderedDict[str, Any]:
        primary_xml_url = jmespath.search(
            expression="""repomd.data[? "@type"=='primary'].location."@href" | [0]""",
            data=repomd_dict,
        )

        primary_xml_url = repo_url / primary_xml_url

        try:
            response = self._stream_url_data(url=primary_xml_url, auth=auth, cert=cert)
        except HTTPError as ex:
            raise HopprPluginError(f"Failed to get primary XML data from {primary_xml_url}") from ex

        data: OrderedDict[str, Any] = OrderedDict()

        # Parse primary XML data to dict
        for chunk in response.iter_content(chunk_size=None):
            data = xmltodict.parse(xml_input=gzip.decompress(chunk), force_list=["package"])

        return data

    def _get_repodata(
        self, repo_url: RepositoryUrl, auth: HTTPBasicAuth | None = None, cert: tuple[str, str] | None = None
    ) -> OrderedDict[str, Any]:
        repomd_url = repo_url / "repodata" / "repomd.xml"

        try:
            response = self._stream_url_data(url=repomd_url, auth=auth, cert=cert)
        except HTTPError as ex:
            raise HopprPluginError(f"Failed to get repository metadata from {repo_url}") from ex

        repomd_dict: OrderedDict[str, Any] = xmltodict.parse(xml_input=response.text, force_list=["data"])

        # Download all metadata files listed in repomd.xml to bundle directory for this repo
        metadata_files = jmespath.search(expression='repomd.data[].location."@href"', data=repomd_dict)
        repodata_dir = self.directory_for(purl_type="rpm", repo_url=str(repo_url), subdir="repodata")

        for metadata_file in metadata_files:
            self.get_logger().debug("Downloading %s", metadata_file, indent_level=2)
            hoppr.net.download_file(url=str(repo_url / metadata_file), dest=str(repodata_dir.parent / metadata_file))

        return repomd_dict

    def _populate_rpm_data(self, repo_url: RepositoryUrl) -> None:
        """Populate `rpm_data` dict for a repository.

        Args:
            repo_url (str): The RPM repository URL
        """
        if type(self).rpm_data.get(f"{repo_url}"):
            return

        self.get_logger().debug("Populating RPM data for repository: %s", repo_url, indent_level=1)

        auth: HTTPBasicAuth | None = None

        if (creds := Credentials.find(f"{repo_url}")) and isinstance(creds.password, SecretStr):
            auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())
            self.repo_configs.append(
                RepoConfig(
                    base_url=[str(repo_url)], username=creds.username, password=creds.password.get_secret_value()
                )
            )
        else:
            self.repo_configs.append(
                RepoConfig(
                    base_url=[str(repo_url)],
                )
            )

        cert = None
        for repo in self.repo_configs:
            for url in repo.base_url:
                if url in f"{repo_url}":
                    if repo.ssl_client_cert and repo.ssl_client_key:
                        cert = (str(repo.ssl_client_cert), str(repo.ssl_client_key))
                    elif repo.username and repo.password:
                        auth = HTTPBasicAuth(username=repo.username, password=repo.password)

        try:
            repomd_dict: OrderedDict[str, Any] = self._get_repodata(repo_url, auth=auth, cert=cert)
            primary_xml_data: OrderedDict[str, Any] = self._get_primary_xml_data(repo_url, repomd_dict, auth, cert)
        except HopprPluginError as ex:
            raise ex

        type(self).rpm_data[f"{repo_url}"] = primary_xml_data

    def _stream_url_data(
        self,
        url: RepositoryUrl,
        auth: HTTPBasicAuth | None = None,
        cert: tuple[str, str] | None = None,
    ) -> Response:
        """Stream download data from specified URL.

        Args:
            url: URL of remote resource to stream.
            auth: Basic authentication if required by URL. Defaults to None.
            cert: Certificate and key to use for request. Defaults to None.

        Raises:
            HTTPError: Failed to download resource after 3 attempts.

        Returns:
            The web request response.
        """
        response = Response()

        for _ in range(self.REQUEST_RETRIES):
            response = requests.get(
                url=f"{url}",
                auth=auth,
                stream=True,
                timeout=self.REQUEST_TIMEOUT,
                proxies=self._get_proxies_for(f"{url}"),
                cert=cert,
            )

            try:
                response.raise_for_status()
                return response
            except HTTPError:
                time.sleep(self.REQUEST_RETRY_INTERVAL)

        raise HTTPError(f"Failed to retrieve data from {url}", response=response)

    @hoppr_process
    @hoppr_rerunner
    def pre_stage_process(self) -> Result:  # sourcery skip: use-named-expression
        """Populate RPM data mapping for repositories."""
        # Get all repository search sequences from all components
        results: list[str] = jmespath.search(
            expression=f"""
                components[*] | [?
                    type != 'operating-system' &&
                    not_null(purl) &&
                    starts_with(purl, 'pkg:rpm')
                ].properties[] | [?
                    name=='{BomProps.COMPONENT_SEARCH_SEQUENCE.value}'
                ].value
            """,
            data=self.context.consolidated_sbom.dict(),
        )

        # Parse and flatten repositories from component search sequence JSON strings, then remove duplicates
        search_repos = hoppr.utils.dedup_list([
            str(repo) for result in results for repo in SearchSequence.parse_raw(result).repositories
        ])

        for repo_url in [RepositoryUrl(url=url) for url in search_repos]:
            self._populate_rpm_data(repo_url)

        # Get all system-managed .repo files and append urls to system repositories
        if self._get_repo_files():
            base_urls = [url for repo in self.repo_configs for url in repo.base_url]
            for url in base_urls:
                type(self).system_repositories.append(url)
                self._populate_rpm_data(RepositoryUrl(url=url))

        return Result.success()

    def _get_repo_files(self) -> Generator[Path, None, None] | None:
        """Return a list of the system repo files and update the plugin repo config."""
        repo_files = None

        if not self.context.strict_repos and os.name != "nt":
            # Set RepoConfig class variables
            RepoConfig.base_arch = self.collector_config.base_arch
            RepoConfig.release_ver = self.collector_config.release_ver
            RepoConfig.vars_dir = self.collector_config.vars_dir

            # Get all system-managed .repo files
            repo_files = Path("/", "etc", "yum.repos.d").glob("*.repo")

            for file in repo_files:
                self.repo_configs.extend(RepoConfigList.parse_file(file))

        return repo_files

    def _validate_hashes(self, package_hash: Hash, comp: Component):
        """Validate generated hash to the RPM package hash.

        Args:
            package_hash: hash pulled from the package file
            comp: Component object whose name will be referenced in case of an error
        Raises:
            HopprPluginError: raise HopprPluginError in case of hash mismatch.
        """
        if all(package_hash != comp_hash for comp_hash in comp.hashes):
            raise HopprPluginError(f"Hash for {comp.name} does not match expected hash.")

    @override
    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        purl = hoppr.utils.get_package_url(comp.purl)

        try:
            download_url, hash_alg, hash_string = self._get_download_url(purl=purl, repo_url=repo_url)
        except HopprPluginError as ex:
            return Result.retry(message=str(ex))

        target_dir = self.directory_for(
            purl_type="rpm",
            repo_url=repo_url,
            subdir=Path(download_url).parent.relative_to(repo_url),
        )

        dest_file = target_dir / Path(download_url).name

        cert = None

        # Obtain cert from local repositories
        repo_files = self._get_repo_files()
        if repo_files is not None:
            for repo in self.repo_configs:
                for url in repo.base_url:
                    if url in f"{repo_url}" and repo.ssl_client_cert and repo.ssl_client_key:
                        cert = (str(repo.ssl_client_cert), str(repo.ssl_client_key))

        try:
            self._download_component(download_url, dest_file, creds, cert)
            self._update_hashes(comp, dest_file)
            self._validate_hashes(
                Hash(
                    alg=next(key for key, value in hoppr.net.HASH_ALG_MAP.items() if value == hash_alg),
                    content=hash_string,
                ),
                comp,
            )
        except (HopprPluginError, ValueError) as ex:
            return Result.fail(message=str(ex))

        self.set_collection_params(comp, repo_url, target_dir)

        return Result.success(return_obj=comp)



"""`validate` subcommand for `hopctl`."""

from __future__ import annotations

import functools
import json
import logging
import os
import sys

from datetime import date, datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, Literal, TypeAlias, cast

from pydantic.utils import deep_update
from rich.console import Group
from rich.live import Live
from rich.rule import Rule
from rich.table import Table
from rich.text import Text
from typer import Argument, BadParameter, CallbackParam, Context, Option, Typer

import hoppr.main
import hoppr.models.validation
import hoppr.net
import hoppr.utils

from hoppr.cli.layout import HopprBasePanel, HopprSbomFilesLayout, console
from hoppr.cli.options import (
    basic_term_option,
    experimental_option,
    log_file_option,
    manifest_file_option,
    sbom_dirs_option,
    sbom_files_option,
    sbom_urls_option,
    verbose_option,
)
from hoppr.logger import HopprLogger
from hoppr.models.licenses import LicenseExpressionItem, LicenseMultipleItem
from hoppr.models.sbom import Component, Metadata, Sbom
from hoppr.models.types import LocalFile, UrlFile
from hoppr.models.validate import (
    ValidateCheckResult,
    ValidateComponentResult,
    ValidateLicenseResult,
    ValidateSbomResult,
)
from hoppr.models.validation.checks import ValidateConfig
from hoppr.result import ResultStatus

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping

    from rich.console import Console, ConsoleOptions, RenderResult

    DictStrAny: TypeAlias = dict[str, Any]

CURRENT_DATE: Final[date] = date.today()
EXPIRATION_DAYS: float = 30

RESULT_MAP: Final[Mapping[ResultStatus, Text]] = {
    ResultStatus.FAIL: Text("\u274c\ufe0f"),
    ResultStatus.SUCCESS: Text.from_markup("[green]\u2714"),
    ResultStatus.WARN: Text("\u26a0\ufe0f"),
}

STRICT_ALL = False
STRICT_LICENSE = False
STRICT_NTIA = False

layout: HopprSbomFilesLayout
logger: HopprLogger

app = Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Validate CycloneDX SBOMs or Hoppr config files",
    no_args_is_help=True,
    rich_markup_mode="rich",
)

LICENSE_FAIL_LOG_LEVEL: int
LICENSE_FAIL_RESULT: ValidateCheckResult
NTIA_FAIL_LOG_LEVEL: int
NTIA_FAIL_RESULT: ValidateCheckResult


class HopprValidateSummary(HopprBasePanel):
    """`hopctl validate sbom` results summary."""

    summary_group = Group()

    def __init__(self, *args, title: str | None = "Summary", log_file: Path, **kwargs) -> None:
        self.log_file = log_file
        self.results: list[ValidateSbomResult] = []

        super().__init__(self.summary_group, title, *args, **kwargs)

    def _component_ntia_row(self, component_results: list[ValidateComponentResult]) -> tuple[str, Text]:
        component_ntia_fails = [cr for cr in component_results if not cr.ntia_fields_result.is_success()]
        ntia_fail_status = RESULT_MAP[NTIA_FAIL_RESULT.status]

        return (
            f"{len(component_ntia_fails)} out of {len(component_results)} components missing minimum NTIA fields",
            ntia_fail_status if component_ntia_fails else RESULT_MAP[ResultStatus.SUCCESS],
        )

    def _license_fields_row(
        self,
        license_results: list[ValidateLicenseResult],
        result_type: Literal["component", "SBOM"],
    ) -> tuple[str, Text]:
        license_field_fails = [lr for lr in license_results if not lr.required_fields.is_success()]

        license_fail_status = RESULT_MAP[LICENSE_FAIL_RESULT.status]

        return (
            f"{len(license_field_fails)} out of {len(license_results)} "
            f"{result_type} licenses missing minimum license fields",
            license_fail_status if license_field_fails else RESULT_MAP[ResultStatus.SUCCESS],
        )

    def _license_expiration_row(
        self,
        license_results: list[ValidateLicenseResult],
        result_type: Literal["component", "SBOM"],
    ) -> tuple[str, Text]:
        license_exp_fails = [lr for lr in license_results if not lr.expiration.is_success()]

        license_fail_status = RESULT_MAP[LICENSE_FAIL_RESULT.status]

        return (
            f"{len(license_exp_fails)} out of {len(license_results)} "
            f"{result_type} licenses expired or expiring within {EXPIRATION_DAYS:n} days",
            license_fail_status if license_exp_fails else RESULT_MAP[ResultStatus.SUCCESS],
        )

    def __rich_console__(self, console_: Console, options: ConsoleOptions) -> RenderResult:
        for sbom_result in self.results:
            component_license_checks = [lr for cr in sbom_result.component_results for lr in cr.license_results]

            self.summary_group.renderables.append(f"Results for {sbom_result.name}:")
            summary_table = Table.grid(padding=(0, 4), pad_edge=True)

            for result in [
                sbom_result.spec_version_result,
                sbom_result.ntia_fields_result,
            ]:
                summary_table.add_row(result.message, RESULT_MAP[result.status])

            # Add SBOM metadata license results summary rows
            summary_table.add_row(*self._license_fields_row(sbom_result.license_results, "SBOM"))
            summary_table.add_row(*self._license_expiration_row(sbom_result.license_results, "SBOM"))

            # Add component NTIA minimum fields and license results summary rows
            summary_table.add_row(*self._component_ntia_row(sbom_result.component_results))
            summary_table.add_row(*self._license_fields_row(component_license_checks, "component"))
            summary_table.add_row(*self._license_expiration_row(component_license_checks, "component"))

            self.summary_group.renderables.append(summary_table)
            self.summary_group.renderables.append(Rule(style="[white]rule.line"))

        self.summary_group.renderables.append(f"See log file {self.log_file} for full results.")

        if hoppr.utils.is_basic_terminal():
            console_.line()
            console_.rule(title="Summary", characters="=", style="[bold]rule.line")
            console_.print(self.summary_group)
            return []

        return super().__rich_console__(console_, options)


class ValidateOutputFormat(str, Enum):
    """Output format for `hopctl validate sbom` when specifying an output file."""

    GITLAB = "gitlab"
    JSON = "json"


def _init_globals(
    log_file: Path,
    verbose: bool,
    expiration_days: float,
    strict: bool,
    strict_license_fields: bool,
    strict_ntia_minimum_fields: bool,
):
    global layout, logger

    logger = HopprLogger(
        name="hopctl-validate",
        filename=str(log_file),
        level=logging.DEBUG if verbose else logging.INFO,
    )

    layout = HopprSbomFilesLayout()

    global EXPIRATION_DAYS, STRICT_ALL, STRICT_LICENSE, STRICT_NTIA

    EXPIRATION_DAYS = expiration_days
    STRICT_ALL = strict
    STRICT_LICENSE = strict_license_fields
    STRICT_NTIA = strict_ntia_minimum_fields

    global LICENSE_FAIL_LOG_LEVEL, LICENSE_FAIL_RESULT, NTIA_FAIL_LOG_LEVEL, NTIA_FAIL_RESULT

    LICENSE_FAIL_LOG_LEVEL = logging.ERROR if STRICT_LICENSE else logging.WARN
    LICENSE_FAIL_RESULT = ValidateCheckResult.fail() if STRICT_LICENSE else ValidateCheckResult.warn()
    NTIA_FAIL_LOG_LEVEL = logging.ERROR if STRICT_NTIA else logging.WARN
    NTIA_FAIL_RESULT = ValidateCheckResult.fail() if STRICT_NTIA else ValidateCheckResult.warn()


def _available_profiles() -> Iterator[str]:
    """Iterate over available validation profiles.

    Returns:
        Iterator of available profiles
    """
    profile_dir = Path(hoppr.models.validation.__file__).parent / "profiles"

    yield from (path.name.removesuffix(".config.yml") for path in profile_dir.glob("*.config.yml"))


def _profile_callback(profile: str | None) -> str:
    """Verifies that the profile given exists or is None.

    Args:
        profile: The parameter value passed in

    Returns:
        The flag that was passed or "default" if not
    """
    if profile not in [None, *_available_profiles()]:
        raise BadParameter(f"Specified profile '{profile}' is invalid. Must be one of {_available_profiles()}")

    return "default" if profile is None else profile


def _config_file_callback(ctx: Context, config_file: Path | None) -> Path:
    """Returns the config file to use for SBOM validation.

    Order of precedence:
      1. User specified path
      2. ~/.config/hoppr.config.yml
      3. ./hoppr.config.yml
      4. default config located at hoppr/models/validation/profiles/default.config.yml

    Args:
        ctx: Click context to update
        config_file: The parameter value passed in

    Returns:
        The path to the config file to use
    """
    profiles_dir = Path(hoppr.models.validation.__file__).parent / "profiles"
    profile = ctx.params.get("profile", "default")

    config_file = cast(Path, config_file or next(
            (
                cfg for cfg in [
                    Path.home() / ".config" / "hoppr.config.yml",
                    Path.cwd() / "hoppr.config.yml",
                ]
                if cfg.exists()
            ),
            profiles_dir / f"{profile}.config.yml",
        ),
    )  # fmt: skip

    # Update merged config with config file values
    merged_config = cast(dict[str, Any], hoppr.utils.load_file(profiles_dir / f"{profile}.config.yml"))
    config_dict = cast(dict[str, Any], hoppr.utils.load_file(config_file))
    merged_config = deep_update(merged_config, config_dict)

    # Load the configuration
    ValidateConfig.parse_obj(merged_config)

    return config_file


def _expiration_days_callback(expiration_days: float) -> float:
    os.environ["HOPPR_EXPIRATION_DAYS"] = str(int(expiration_days))

    return expiration_days


def _set_all_strict(ctx: Context, param: CallbackParam, value: bool) -> bool:
    """Set all strict flags to `True`.

    Args:
        ctx (Context): Click context to update
        param (CallbackParam): Typer metadata for the parameter
        value (bool): The parameter value

    Returns:
        bool: The flag that was passed, unmodified
    """
    if param.name and value:
        ctx.params[param.name] = True
        ctx.params["strict_ntia_minimum_fields"] = True
        ctx.params["strict_license_fields"] = True

    return value


def _check_fields(
    obj: object | None,
    *field_names: str,
    fail_log_level: int = logging.WARN,
    fail_result: ValidateCheckResult = ValidateCheckResult.warn(),  # noqa: B008
    indent_level: int = 1,
) -> ValidateCheckResult:
    result = ValidateCheckResult.success()

    if isinstance(obj, dict):
        field_values = (obj.get(field_name) for field_name in field_names)
    else:
        field_values = (getattr(obj, field_name, None) for field_name in field_names)

    if obj is None or not any(field_values):
        missing_fields = ", ".join(f"`{name}`" for name in field_names)
        logger.log(
            fail_log_level,
            "Missing %s field",
            missing_fields,
            indent_level=indent_level,
        )

        result.merge(fail_result)

    return result


def _check_license_expiration(
    license_: LicenseMultipleItem | LicenseExpressionItem,
) -> ValidateCheckResult:
    if isinstance(license_, LicenseExpressionItem) or not license_.license.licensing:
        return LICENSE_FAIL_RESULT

    try:
        # Raises `ValueError` if `expiration` is `None`
        expiration = datetime.fromisoformat(str(license_.license.licensing.expiration))
        if expiration.date() < CURRENT_DATE + timedelta(days=EXPIRATION_DAYS):
            raise ValueError
    except ValueError:
        logger.log(
            LICENSE_FAIL_LOG_LEVEL,
            "License expired or expiring within %d days",
            EXPIRATION_DAYS,
            indent_level=3,
        )

        return LICENSE_FAIL_RESULT

    return ValidateCheckResult.success()


def _check_license_fields(
    license_: LicenseMultipleItem | LicenseExpressionItem,
) -> ValidateCheckResult:
    required_fields_result = ValidateCheckResult.success()

    if isinstance(license_, LicenseExpressionItem):  # pragma no cover
        required_fields_result.merge(LICENSE_FAIL_RESULT)
        return required_fields_result

    check_license_fields = functools.partial(
        _check_fields,
        fail_log_level=LICENSE_FAIL_LOG_LEVEL,
        fail_result=LICENSE_FAIL_RESULT,
        indent_level=3,
    )

    required_fields_result.merge(check_license_fields(license_.license, "id", "name"))
    required_fields_result.merge(check_license_fields(license_.license, "licensing"))
    return required_fields_result


def _check_spec_version(sbom_dict: dict[str, Any]) -> ValidateCheckResult:
    result = ValidateCheckResult.success()

    match sbom_dict.get("specVersion"):
        case "1.2" | None:
            result.merge(ValidateCheckResult.fail())
        case "1.3" | "1.4":
            result.merge(ValidateCheckResult.fail() if STRICT_ALL else ValidateCheckResult.warn())

    return result


def _validate_component(component: Component) -> ValidateComponentResult:
    component_id = "@".join(filter(None, [component.name, component.version]))
    comp_result = ValidateComponentResult(component_id=component_id)

    logger.info("Validating component: %s", component_id, indent_level=1)

    check_component_fields = functools.partial(
        _check_fields,
        fail_log_level=NTIA_FAIL_LOG_LEVEL,
        fail_result=NTIA_FAIL_RESULT,
        indent_level=2,
    )

    # Validate minimum required NTIA fields for component
    comp_result.ntia_fields_result.merge(check_component_fields(component, "supplier"))
    comp_result.ntia_fields_result.merge(check_component_fields(component, "name"))
    comp_result.ntia_fields_result.merge(check_component_fields(component, "version"))
    comp_result.ntia_fields_result.merge(check_component_fields(component, "cpe", "purl", "swid"))
    comp_result.result.merge(comp_result.ntia_fields_result)

    license_results = _validate_licenses(component)

    for license_result in license_results:
        comp_result.license_results.append(license_result)

        # Merge license result into final result for this SBOM
        comp_result.result.merge(license_result.required_fields)
        comp_result.result.merge(license_result.expiration)

    return comp_result


def _validate_licenses(obj: Metadata | Component | None) -> list[ValidateLicenseResult]:
    license_results: list[ValidateLicenseResult] = []

    if not obj or not obj.licenses:
        logger.log(LICENSE_FAIL_LOG_LEVEL, "Missing `licenses` field", indent_level=2)

        license_result = ValidateLicenseResult(license_id="Missing license data")
        license_result.required_fields.merge(LICENSE_FAIL_RESULT)
        license_result.expiration.merge(LICENSE_FAIL_RESULT)
        license_results.append(license_result)

        return license_results

    for license_ in obj.licenses:
        license_id = (
            license_.expression
            if isinstance(license_, LicenseExpressionItem)
            else str(license_.license.id or license_.license.name)
        )

        logger.info("Validating license: %s", license_id, indent_level=2)

        license_result = ValidateLicenseResult(license_id=license_id)
        license_result.required_fields.merge(_check_license_fields(license_))

        # Validate license expiration not within specified number of days
        license_result.expiration.merge(_check_license_expiration(license_))

        license_results.append(license_result)

    return license_results


def _validate_sbom(name: str, sbom_data: dict[str, Any]) -> ValidateSbomResult:
    sbom_result = ValidateSbomResult(name=name)

    # Check raw `specVersion` and NTIA minimum fields values before parsing data as Sbom object
    sbom_result.spec_version_result.merge(_check_spec_version(sbom_data))

    check_sbom_fields = functools.partial(
        _check_fields,
        fail_log_level=NTIA_FAIL_LOG_LEVEL,
        fail_result=NTIA_FAIL_RESULT,
        indent_level=1,
    )

    sbom_result.ntia_fields_result.merge(check_sbom_fields(sbom_data, "metadata"))
    sbom_result.ntia_fields_result.merge(check_sbom_fields(sbom_data.get("metadata"), "authors", "tools"))
    sbom_result.ntia_fields_result.merge(check_sbom_fields(sbom_data.get("metadata"), "timestamp"))
    sbom_result.result.merge(sbom_result.ntia_fields_result)

    logger.info("Validating SBOM licenses", indent_level=1)

    # Parse SBOM data and run all validation checks
    sbom_ = Sbom.load(sbom_data)
    sbom_result.license_results = _validate_licenses(sbom_.metadata)

    # Merge license results into final result for this SBOM
    for license_result in sbom_result.license_results:
        sbom_result.result.merge(license_result.required_fields)
        sbom_result.result.merge(license_result.expiration)

    # Skip validation for components with type of `operating-system`
    for component in [comp for comp in sbom_.components if str(comp.type) != "operating-system"]:
        component_result = _validate_component(component)
        sbom_result.component_results.append(component_result)

        # Merge component NTIA fields result into final result for this SBOM
        sbom_result.ntia_fields_result.merge(component_result.ntia_fields_result)

        # Merge component result into final result for this SBOM
        sbom_result.result.merge(component_result.result)

    return sbom_result


def _create_sbom_data_map(manifest_file: Path, sbom_files: list[Path], sbom_urls: list[str]) -> dict[str, DictStrAny]:
    """Create mapping of SBOM file paths to file contents.

    Args:
        manifest_file: Path to manifest file
        sbom_files: A list of paths for sbom files
        sbom_urls: A list of paths for sbom URLs

    Returns:
        Mapping of SBOM file paths to file contents
    """
    if manifest_file:
        # Get all SBOMs that were included in the manifest file
        sbom_files.extend(file.local for file in Sbom.loaded_sboms if isinstance(file, LocalFile))
        sbom_urls.extend(file.url for file in Sbom.loaded_sboms if isinstance(file, UrlFile))

    sbom_data_map: dict[str, dict[str, Any]] = {}

    for sbom_file in sbom_files:
        sbom_dict = hoppr.utils.load_file(sbom_file)
        if not isinstance(sbom_dict, dict):
            raise TypeError("SBOM file was not loaded as dictionary")

        sbom_data_map[str(sbom_file)] = sbom_dict
        layout.add_job(description=sbom_file.name)

    for sbom_url in sbom_urls:
        sbom_dict = hoppr.net.load_url(sbom_url)
        if not isinstance(sbom_dict, dict):
            raise TypeError("SBOM URL was not loaded as dictionary")

        sbom_data_map[sbom_url] = sbom_dict
        layout.add_job(description=Path(sbom_url).name)

    return sbom_data_map


@app.command(no_args_is_help=True)
def sbom(
    manifest_file: Path = manifest_file_option,
    sbom_files: list[Path] = sbom_files_option,
    sbom_dirs: list[Path] = sbom_dirs_option,
    sbom_urls: list[str] = sbom_urls_option,
    strict: bool = Option(
        False,
        "-S",
        "--strict",
        callback=_set_all_strict,
        help="Enable all strict validation options",
        show_default=False,
    ),
    strict_ntia_minimum_fields: bool = Option(
        False,
        "-n",
        "--strict-ntia-minimum-fields",
        help="Raise error if minimum fields recommended by NTIA are not set",
        is_eager=True,
        show_default=False,
    ),
    strict_license_fields: bool = Option(
        False,
        "-L",
        "--strict-license-fields",
        help="Raise error if SBOM license or SBOM/component license expiration fields are not set",
        is_eager=True,
        show_default=False,
    ),
    profile: str = Option(
        None,
        "-p",
        "--profile",
        callback=_profile_callback,
        help="Profile of configuration presets to apply [bold](supersedes [cyan]--strict*[/] flags)[/]",
        metavar=f"[{'|'.join(_available_profiles())}]",
        rich_help_panel="Experimental",
        show_default=False,
    ),
    config_file: Path = Option(
        None,
        "-c",
        "--config",
        callback=_config_file_callback,
        dir_okay=False,
        envvar="HOPPR_CONFIG",
        exists=True,
        help="Path to a validation config file",
        rich_help_panel="Experimental",
        show_default=False,
    ),
    expiration_days: float = Option(
        30,
        "-e",
        "--expiration-days",
        callback=_expiration_days_callback,
        help="Number of days allowed by license expiration check",
        show_default=False,
    ),
    output_format: ValidateOutputFormat = Option(
        ValidateOutputFormat.JSON,
        "-f",
        "--output-format",
        help=r"Format for output file [dim]\[default: json]",
        show_default=False,
    ),
    output_file: Path = Option(
        None,
        "-o",
        "--output-file",
        dir_okay=False,
        exists=False,
        help="Path to output file",
        resolve_path=True,
        show_default=False,
    ),
    basic_term: bool = basic_term_option,
    log_file: Path = log_file_option,
    verbose: bool = verbose_option,
    experimental: bool = experimental_option,
):  # pragma: no cover
    """Validate SBOM file(s)."""
    if experimental:
        from hoppr.cli.experimental.validate import _validate_experimental

        _validate_experimental(manifest_file, sbom_files, sbom_urls, output_file, log_file, verbose)

    _init_globals(
        log_file,
        verbose,
        expiration_days,
        strict,
        strict_license_fields,
        strict_ntia_minimum_fields,
    )

    live_display = Live(layout, console=console, refresh_per_second=10)

    summary_panel = HopprValidateSummary(log_file=log_file)

    result = ValidateCheckResult.success()

    sbom_results: list[ValidateSbomResult] = []
    sbom_data_map = _create_sbom_data_map(manifest_file, sbom_files, sbom_urls)

    if not hoppr.utils.is_basic_terminal():
        live_display.start(refresh=True)

    for name, sbom_data in sbom_data_map.items():
        msg = f"Validating {name}..."
        logger.info(msg)
        layout.print(msg)

        sbom_result = _validate_sbom(name, sbom_data)
        summary_panel.results.append(sbom_result)
        result.merge(sbom_result.result)
        sbom_results.append(sbom_result)

        layout.update_job(name=Path(name).name, status=result.status.name)
        layout.stop_job(Path(name).name)
        layout.advance_job(Path(name).name)

    if not hoppr.utils.is_basic_terminal():
        live_display.stop()

    console.print("\n", summary_panel)

    if output_file:
        # Can be extended in the future to write other types of reports
        output_data = json.loads(f"[{', '.join([sbom_result.json() for sbom_result in sbom_results])}]")
        output_file.write_text(data=json.dumps(output_data, indent=2), encoding="utf-8")

    sys.exit(1 if result.is_fail() else 0)


@app.command(no_args_is_help=True)
def config(
    input_files: list[Path] = Argument(
        ...,
        dir_okay=False,
        exists=True,
        help="Path to manifest file",
        resolve_path=True,
        show_default=False,
    ),
    credentials_file: Path = Option(
        None,
        "-c",
        "--credentials",
        help="Specify credentials config for services",
        envvar="HOPPR_CREDS_CONFIG",
        show_default=False,
    ),
    transfer_file: Path = Option(
        "transfer.yml",
        "-t",
        "--transfer",
        help="Specify transfer config",
        envvar="HOPPR_TRANSFER_CONFIG",
    ),
):  # pragma: no cover
    """Validate Hoppr manifest, transfer, and credentials file(s)."""
    hoppr.main.validate(input_files, credentials_file, transfer_file)



"""Base model for Hoppr config files."""

from __future__ import annotations

import functools
import io
import json
import sys

from collections.abc import Callable, MutableMapping
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, ClassVar, Literal, TypeGuard, TypeVar

import hoppr_cyclonedx_models.base as cdx_base
import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx

from pydantic import BaseConfig, BaseModel, Extra, Field, Protocol, create_model, validator
from rich.markdown import Markdown
from ruamel.yaml import YAML
from typing_extensions import Self

import hoppr.utils

if TYPE_CHECKING:
    from pydantic.types import StrBytes
    from pydantic.typing import AbstractSetIntStr, MappingIntStrAny
    from rich.console import Console, ConsoleOptions, RenderResult
    from rich.repr import RichReprResult

    YamlStyle = Literal["", '"', "'", "|", ">"] | None

AnyCycloneDXModel = TypeVar("AnyCycloneDXModel", bound="CycloneDXBaseModel")
ExtendedProto = Protocol | Literal["yaml", "yml"]
UniqueIDMap = Annotated[MutableMapping[str, AnyCycloneDXModel], Field(default=...)]

__all__: list[str] = []


class HopprBaseModel(BaseModel):
    """Base Hoppr data model."""

    class Config(BaseConfig):
        """Config options for HopprBaseModel."""

        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        extra = Extra.forbid
        use_enum_values = True

    def __eq__(self, other: object) -> bool:
        return hash(self) == hash(other)

    def __hash__(self) -> int:
        """Define to test equality or uniqueness between objects."""
        return hash(repr(self))

    def __repr__(self) -> str:
        attributes = ", ".join(f"{key}={value!r}" for key, value in self.__dict__.items() if value)
        return f"{type(self).__name__}({attributes})"

    def __rich_repr__(self) -> RichReprResult:
        yield from [(key, value) for key, value in self.__dict__.items() if value]

    @classmethod
    def parse_file(
        cls,
        path: str | Path,
        *,
        content_type: str | None = None,
        encoding: str = "utf-8",
        proto: ExtendedProto | None = None,
        allow_pickle: bool = False,
    ) -> Self:
        path = Path(path)

        return cls.parse_raw(
            path.read_text(),
            allow_pickle=allow_pickle,
            content_type=content_type,
            encoding=encoding,
            proto=proto,
        )

    @classmethod
    def parse_raw(
        cls,
        b: StrBytes,
        *,
        content_type: str | None = None,
        encoding: str = "utf-8",
        proto: ExtendedProto | None = None,
        allow_pickle: bool = False,
    ) -> Self:
        return cls.parse_obj(hoppr.utils.load_string(b if isinstance(b, str) else b.decode()))

    def yaml(
        self,
        *,
        include: AbstractSetIntStr | MappingIntStrAny | None = None,
        exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
        by_alias: bool = True,
        skip_defaults: bool | None = None,
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = True,
        sort_keys: bool = False,
        default_flow_style: bool | None = False,
        default_style: YamlStyle = None,
        indent: bool | None = True,
        encoding: str | None = None,
        **kwargs,
    ) -> str:
        """Generate a YAML representation of the model instance.

        Args:
            include: Fields to include. See `BaseModel.dict()`.
            exclude: Fields to exclude. See `BaseModel.dict()`.
            by_alias: Whether to use aliases instead of declared names. Defaults to False.
            skip_defaults: See `BaseModel.dict()`.
            exclude_unset: See `BaseModel.dict()`.
            exclude_defaults: See `BaseModel.dict()`.
            exclude_none: See `BaseModel.dict()`.
            sort_keys: If True, will sort the keys in alphanumeric order for dictionaries.
                Defaults to False, which will dump in the field definition order.
            default_flow_style: Whether to use the "flow" style in the dumper. Defaults to False,
                which uses the "block" style (probably the most familiar to users).
            default_style: This is the default style for quoting strings, used by `ruamel.yaml` dumper.
                One of: {None, "", "'", '"', "|", ">"}.
                Defaults to None, which varies the style based on line length.
            indent: Additional arguments for the dumper.
            encoding: Additional arguments for the dumper.
            kwargs: Additional arguments for the dumper.

        Returns:
            The encoded YAML string.
        """
        # Serialize to JSON first rather than calling self.dict()
        model_json = self.json(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

        data = json.loads(model_json)

        stream = io.StringIO()
        stream.write("---\n")

        yaml_writer = YAML(typ="safe", pure=True)
        yaml_writer.default_flow_style = default_flow_style
        yaml_writer.default_style = default_style  # type: ignore[assignment]
        yaml_writer.encoding = encoding or "utf-8"
        yaml_writer.line_break = "\n"  # type: ignore[assignment]
        yaml_writer.map_indent = 2 if indent else None
        yaml_writer.sequence_dash_offset = 2 if indent else 0
        yaml_writer.sequence_indent = 4 if indent else None
        yaml_writer.sort_base_mapping_type_on_output = sort_keys  # type: ignore[assignment]
        yaml_writer.width = 120

        for key, value in data.items():
            yaml_writer.dump(data={key: value}, stream=stream)
            stream.write("\n")

        # Remove final extra newline character
        stream.seek(stream.tell() - 1)
        stream.truncate()

        return stream.getvalue().replace("'", '"')


class HopprMetadata(BaseModel):
    """Metadata data model."""

    name: str
    version: str | int
    description: str


class HopprBaseSchemaModel(HopprBaseModel):
    """Base Hoppr config file schema model."""

    kind: Literal["Credentials", "Manifest", "Transfer"] = Field(description="Data model/schema kind")
    metadata: HopprMetadata | None = Field(default=None, description="Metadata for the file")
    schema_version: str = Field(default="v1", alias="schemaVersion", title="Schema Version")

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:  # pragma: no cover
        yield Markdown(
            markup="\n".join([
                "```yaml",
                self.yaml(),
                "```",
            ]).strip("\n"),
            code_theme="github-dark",
        )

    @validator("kind", allow_reuse=True, pre=True)
    @classmethod
    def validate_kind(cls, kind: str) -> str:
        """Return supplied `kind` value with only first letter capitalized."""
        return kind.capitalize()


class CycloneDXBaseModel(
    cdx_base.CycloneDXBaseModel,
    allow_population_by_field_name=True,
    arbitrary_types_allowed=True,
    extra=Extra.allow,
    use_enum_values=True,
):
    """Base CycloneDX data model."""

    # Defining as ClassVar to allow dynamic model creation using custom root types
    deep_merge: ClassVar[bool] = False
    flatten: ClassVar[bool] = False
    observers: ClassVar[dict[object, Callable]] = {}
    unique_id_map: ClassVar[UniqueIDMap] = {}

    def __eq__(self, other: object) -> bool:
        return hash(self) == hash(other) if isinstance(other, type(self)) else False

    def __hash__(self) -> int:
        return hash(self.unique_id_callback())

    def __init__(self, **data):
        super().__init__(**data)

        unique_id = self.unique_id_callback()
        type(self).unique_id_map[unique_id] = self

    def _has_field(self, field_name: str) -> bool:
        return hasattr(self, field_name) and (getattr(self, field_name) or field_name in list(self.__fields_set__))

    @staticmethod
    def _is_model_list_field(field: object) -> TypeGuard[list[CycloneDXBaseModel]]:
        """Checks if the provided field is a list of the specified type."""
        return isinstance(field, list) and bool(field) and all(isinstance(item, CycloneDXBaseModel) for item in field)

    def _merge_field(self, target_field_name: str, source_field: object) -> None:
        """Merges `source_field` into the field referenced by `target_field_name`."""
        merged_field = getattr(self, target_field_name)
        field_type = type(merged_field)

        if not self._has_field(target_field_name):
            setattr(self, target_field_name, source_field)
            return

        if isinstance(field_type, type(BaseModel)) and isinstance(source_field, field_type):
            merged_field = CycloneDXBaseModel.create(model=merged_field)
            source_field = type(merged_field).create(model=source_field)
            merged_field.merge(source_field)
        elif self._is_model_list_field(merged_field) or self._is_model_list_field(source_field):
            merged_field = hoppr.utils.dedup_list([CycloneDXBaseModel.create(model=_field) for _field in merged_field])
            source_field = [CycloneDXBaseModel.create(model=_field) for _field in source_field or []]  # type: ignore[attr-defined]
            self._merge_field_items(merged_field, source_field)

        setattr(self, target_field_name, merged_field)

    @staticmethod
    def _merge_field_items(target_field: list[CycloneDXBaseModel], source_field: list[CycloneDXBaseModel]) -> None:
        """Merges the items from `source_field` into `target_field`."""
        for source_item in source_field:
            if source_item in target_field:
                merged_item = target_field[target_field.index(source_item)]
                merged_item.merge(source_item)
            else:
                target_field.append(source_item)

    def unique_id_callback(self) -> str:
        """Default callback method to get a model object's unique ID."""
        try:
            callback = {
                "Advisory": lambda obj: obj.url,
                "Annotations": lambda obj: obj.bom_ref or obj.text,
                "Command": lambda obj: obj.executed or repr(self),
                "Commit": lambda obj: obj.uid,
                "Component": lambda obj: obj.bom_ref,
                "ComponentData": lambda obj: obj.bom_ref or repr(self),
                "Copyright": lambda obj: obj.text,
                "Dependency": lambda obj: obj.ref,
                "ExternalReference": lambda obj: f"{obj.type}-{obj.url}",
                "Formula": lambda obj: obj.bom_ref or repr(self),
                "License": lambda obj: obj.id or obj.name or repr(self),
                "LicenseChoice": lambda obj: obj.license or obj.expression or repr(self),
                "Note": lambda obj: obj.text,
                "Occurrence": lambda obj: obj.bom_ref or repr(self),
                "Reference": lambda obj: obj.id,
                "Sbom": lambda obj: obj.serialNumber,
                "Service": lambda obj: obj.bom_ref or f"{obj.name}-{obj.version}",
                "Signer": lambda obj: obj.value,
                "Task": lambda obj: obj.bom_ref or obj.uid,
                "Vulnerability": lambda obj: obj.id or repr(self),
                "Workflow": lambda obj: obj.bom_ref or obj.uid,
                "Workspace": lambda obj: obj.bom_ref or obj.uid,
            }[type(self).__name__]

            return callback(self)
        except KeyError:
            return repr(self)

    @classmethod
    def create(cls, model: cdx_base.CycloneDXBaseModel) -> Self:
        """Update a BaseModel object with CycloneDXBaseModel attributes and methods.

        Args:
            model (cdx_base.CycloneDXBaseModel): The `hoppr-cyclonedx-models` object to update

        Returns:
            AnyCycloneDXModel: The updated BaseModel object
        """
        model_cls = cls.make_model(name=type(model).__name__)
        return model_cls(**model.dict(by_alias=True, exclude_none=True, exclude_unset=True))

    @classmethod
    @functools.cache
    def make_model(cls, name: str) -> type[Self]:
        """Dynamically create a model class suitable for merging.

        Args:
            name (str): Name of the existing model

        Returns:
            type[AnyCycloneDXModel]: The generated model class
        """
        # Return explicitly defined models directly
        members = {
            **sys.modules[f"{__package__}.affect"].__dict__,
            **sys.modules[f"{__package__}.annotations"].__dict__,
            **sys.modules[f"{__package__}.licenses"].__dict__,
            **sys.modules[f"{__package__}.sbom"].__dict__,
        }

        if model_cls := members.get(name):
            return model_cls

        model_cls = cdx.__dict__[name]

        merge_model = create_model(model_cls.__name__, __base__=(cls, model_cls), __module__=__name__)

        # Set model's pydantic `Config` class and `__hash__` method
        merge_model.__config__ = cls.__config__

        # Add `unique_id_map` class attribute for caching model objects
        merge_model.__class_vars__.add("unique_id_map")
        merge_model.__annotations__["unique_id_map"] = "ClassVar[UniqueIDMap]"
        merge_model.unique_id_map = {}

        # Add updated model to current module and make importable from other modules
        setattr(sys.modules[__name__], merge_model.__name__, merge_model)
        global __all__
        __all__ += [merge_model.__name__]

        merge_model.update_forward_refs()

        return merge_model

    @classmethod
    def find(cls, unique_id: str) -> Self | None:
        """Look up model object by its unique ID string.

        Args:
            unique_id (str): Unique ID string to look up

        Returns:
            AnyCycloneDXModel | None: Model object if found, otherwise None
        """
        return cls.unique_id_map.get(unique_id)

    def merge(self, other: CycloneDXBaseModel) -> None:
        """Merge model instance of same type into self.

        Args:
            other (CycloneDXBaseModel): Model object to merge
        """
        if (self_type := type(self).__name__) != (other_type := type(other).__name__):
            raise TypeError(f"Type '{other_type}' cannot be merged into '{self_type}'")

        self.notify(data=f"  Merging '{type(self).__qualname__}' attributes...")

        for field_name in self.__fields__:
            self.notify(data=f"    Merging field '{type(self).__qualname__}.{field_name}'...")

            if (source_field := getattr(other, field_name, None)) is None:
                continue

            self._merge_field(field_name, source_field)

    def notify(self, data: str) -> None:
        """Call the callback function for all registered subscribers."""
        for callback in self.observers.values():
            callback(data)

    def subscribe(self, observer: object, callback: Callable) -> None:
        """Register an observer."""
        self.observers[observer] = callback

    def unsubscribe(self, observer: object) -> None:
        """Unregister an observer."""
        self.observers.pop(observer, None)



"""Collector plugin for maven artifacts."""

from __future__ import annotations

import importlib
import os
import warnings

from copy import deepcopy
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any

import jmespath
import xmltodict

from typing_extensions import override

import hoppr.net
import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.exceptions import HopprExperimentalWarning, HopprPluginError
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from collections import OrderedDict

    from packageurl import PackageURL

    from hoppr.models import HopprContext
    from hoppr.models.credentials import CredentialRequiredService
    from hoppr.models.sbom import Component


warnings.warn(
    message="This Maven collector plugin is experimental; use at your own risk.",
    category=HopprExperimentalWarning,
    stacklevel=1,
)


class CollectMavenPlugin(
    SerialCollectorPlugin,
    supported_purl_types=["maven"],
    products=["maven/*"],
    system_repositories=["https://repo.maven.apache.org/maven2"],
):
    """Collector plugin for maven artifacts (EXPERIMENTAL)."""

    def get_version(self) -> str:  # noqa: D102
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        system_settings_file = Path("/") / "etc" / "maven" / "settings.xml"
        user_settings_file = Path.home() / ".m2" / "settings.xml"

        if not self.context.strict_repos:
            # Identify system repositories
            for settings_file in [system_settings_file, user_settings_file]:
                if settings_file.is_file():
                    settings_dict: OrderedDict[str, Any] = xmltodict.parse(
                        settings_file.read_text(encoding="utf-8"),
                        encoding="utf-8",
                        force_list={"profile", "repository"},
                    )

                    repo_urls: list[str] = jmespath.search(
                        expression="settings.profiles.profile[].repositories.repository[].url", data=settings_dict
                    )

                    self.system_repositories.extend(
                        repo for repo in repo_urls or []
                        if repo not in self.system_repositories
                    )  # fmt: skip

    def update_hashes(self, comp: Component, package_path: Path) -> Result:
        """Update the hashes of a component with hashes generated from its package file.

        Args:
            comp: Component object whose hashes will be updated.
            package_path: Path of the downloaded package file for the provided component.

        Raises:
            HopprPluginError: If an existing hash doesn't match.
        """
        generated_hashes = self._get_artifact_hashes(package_path)
        try:
            comp.update_hashes(generated_hashes)
        except ValueError as ex:
            raise HopprPluginError(f"Hash for {comp.name} does not match expected hash.") from ex

        return Result.success()

    def _check_artifact_hash(
        self,
        download_url: str,
        dest_file: str,
        creds: CredentialRequiredService | None = None,
    ) -> Result:
        with NamedTemporaryFile(mode="w+", encoding="utf-8", delete=False) as hash_file:
            # Download SHA1 hash file for artifact
            response = hoppr.net.download_file(
                url=f"{download_url}.sha1",
                dest=hash_file.name,
                creds=creds,
                proxies=self._get_proxies_for(download_url),
            )

            result = Result.from_http_response(response=response)
            hash_string = Path(hash_file.name).read_text(encoding="utf-8").strip().lower()
            hash_file.flush()
            hash_file.close()
            Path(hash_file.name).unlink()

        if (computed_hash := hoppr.net.get_file_hash(artifact=dest_file)) != hash_string:
            result.merge(Result.fail(message=f"SHA1 hash for {Path(dest_file).name} does not match expected hash."))
            self.get_logger().debug("Computed hash: %s, expected: %s", computed_hash, hash_string, indent_level=2)

        return result

    # Helper function to combine a classifier, if present, with the artifact version
    def _set_classifier(self, purl: PackageURL) -> str:
        artifact_type = purl.qualifiers.get("type", "jar")
        artifact_classifier = purl.qualifiers.get("classifier")

        return "-".join(filter(None, [purl.version, "" if artifact_type == "pom" else artifact_classifier]))

    def _get_maven_component(
        self,
        purl: PackageURL,
        repo_url: str,
        target_dir: Path,
        creds: CredentialRequiredService | None = None,
    ) -> Result:
        artifact_path = "/".join((purl.namespace or "").split("."))
        artifact_type = purl.qualifiers.get("type", "jar")

        version_qualifiers = self._set_classifier(purl)
        artifact_file = f"{purl.name}-{version_qualifiers}.{artifact_type}"

        download_url = str(RepositoryUrl(url=repo_url) / artifact_path / purl.name / purl.version / artifact_file)

        # Reconstructs the file name to match what is expected by downstream plugins
        file_name = f"{purl.name}_{version_qualifiers}.{artifact_type}"
        dest_file = f"{target_dir / file_name}"

        self.get_logger().info(msg=f"source: {download_url}", indent_level=3)
        self.get_logger().info(msg=f"destination: {dest_file}", indent_level=3)

        response = hoppr.net.download_file(
            url=download_url,
            dest=dest_file,
            creds=creds,
            proxies=self._get_proxies_for(download_url),
        )

        if not (result := Result.from_http_response(response=response)).is_success():
            return result

        result.merge(self._check_artifact_hash(download_url, dest_file, creds))

        return result

    @override
    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        purl = hoppr.utils.get_package_url(comp.purl)
        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)

        artifact_type = purl.qualifiers.get("type", "jar")

        version_qualifiers = self._set_classifier(purl)

        # Reconstructs the file name to match what is expected by downstream plugins
        file_name = f"{purl.name}_{version_qualifiers}.{artifact_type}"
        package_out_path = target_dir / file_name

        self.get_logger().info(msg="Downloading Maven artifact:", indent_level=2)
        result = self._get_maven_component(purl, repo_url, target_dir, creds)

        if not result.is_success():
            msg = f"Failed to download Maven artifact {comp.purl}, {result.message}"
            self.get_logger().error(msg=msg, indent_level=2)
            return Result.fail(message=msg)

        if purl.qualifiers.get("type") != "pom":
            purl_copy = deepcopy(purl)
            purl_copy.qualifiers["type"] = "pom"
            artifact_type_copy = "pom"

            # Don't need to run the _set_classifier function as it has to be a pom here
            file_name = f"{purl.name}_{purl.version}.{artifact_type_copy}"

            self.get_logger().info(msg="Downloading pom for Maven artifact:", indent_level=2)
            result = self._get_maven_component(purl_copy, repo_url, target_dir, creds)

            if not result.is_success():
                msg = f"Failed to download pom for Maven artifact {comp.purl}, {result.message}"
                self.get_logger().error(msg=msg, indent_level=2)
                return Result.fail(message=msg)

        self.update_hashes(comp, Path(package_out_path))

        self.set_collection_params(comp, repo_url, package_out_path)

        return Result.success(return_obj=comp)


if not os.getenv("HOPPR_EXPERIMENTAL"):
    warnings.warn(
        message="Either the --experimental flag or the environment variable "
        "HOPPR_EXPERIMENTAL must be set in order to use this plugin.",
        category=HopprExperimentalWarning,
        stacklevel=1,
    )

    module = importlib.import_module(name="hoppr.core_plugins.collect_maven_plugin")
    CollectMavenPlugin = module.CollectMavenPlugin  # type: ignore[misc]



"""Collector plugin for Cargo packages."""

from __future__ import annotations

from typing import TYPE_CHECKING

import jmespath
import requests as requests

from pydantic import SecretStr
from requests import HTTPError
from requests.auth import HTTPBasicAuth
from typing_extensions import override

import hoppr.net
import hoppr.utils

from hoppr import Hash, __version__, cdx
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.exceptions import HopprPluginError
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from pathlib import Path

    from packageurl import PackageURL

    from hoppr.models.credentials import CredentialRequiredService
    from hoppr.models.sbom import Component


class CollectCargoPlugin(
    SerialCollectorPlugin,
    products=["cargo/*"],
    supported_purl_types=["cargo"],
    system_repositories=["https://crates.io/api/v1/crates"],
):
    """Collector plugin for Cargo packages."""

    def get_version(self) -> str:  # noqa: D102
        return __version__

    def _download_package(
        self, purl: str | None, download_url: str, dest_file: Path, creds: CredentialRequiredService | None
    ):
        """Downloads a cargo package from a specific URL.

        Args:
            purl: Used to identify the package in the logs.
            download_url: Where the package will be downloaded from.
            dest_file: Where the package will be downloaded to.
            creds: Credentials used for the HTTP request.

        Raises:
            HopprPluginError: If package download fails.
        """
        response = hoppr.net.download_file(
            url=download_url,
            dest=str(dest_file),
            creds=creds,
            proxies=self._get_proxies_for(download_url),
            timeout=self.process_timeout,
        )

        try:
            response.raise_for_status()
        except HTTPError as ex:
            msg = f"Failed to download Cargo package {purl}, status_code={response.status_code}"
            self.get_logger().error(msg=msg, indent_level=2)
            raise HopprPluginError(msg) from ex

    def _get_package_checksum(self, metadata_url: str, creds: CredentialRequiredService | None) -> Hash:
        """Gets the package checksum from the metadata at the specified URL.

        Args:
            metadata_url: URL to the JSON document containing the package metadata.
            creds: Credentials used for the HTTP request.

        Returns:
            The SHA256 hash contained in the metadata.

        Raises:
            HopprPluginError: If download fails.
            ValueError: If the downloaded file does not contain the checksum.
        """
        authentication = None
        if creds is not None and isinstance(creds.password, SecretStr):
            authentication = HTTPBasicAuth(creds.username, creds.password.get_secret_value())

        package_metadata_response = requests.get(
            url=metadata_url,
            auth=authentication,
            timeout=300,
            proxies=self._get_proxies_for(metadata_url),
        )

        try:
            package_metadata_response.raise_for_status()
        except HTTPError as ex:
            msg = (
                f"Failed to download Cargo package metadata from {metadata_url}, "
                f"status_code={package_metadata_response.status_code}"
            )
            raise HopprPluginError(msg) from ex

        checksum = jmespath.search(expression="version.checksum", data=package_metadata_response.json())

        if checksum is None:
            raise ValueError("Package metadata is missing package checksum.")

        return Hash(alg=cdx.HashAlg.SHA_256, content=checksum)

    def _get_download_urls(self, purl: PackageURL, repo_url: str) -> tuple[str, str]:
        package_metadata_url = RepositoryUrl(url=repo_url) / purl.name / (purl.version or "")
        package_download_url = package_metadata_url / "download"
        return str(package_metadata_url), str(package_download_url)

    def _update_hashes(self, comp: Component, package_path: Path):
        """Update the hashes of a component with hashes generated from its package file.

        Args:
            comp: Component object whose hashes will be updated.
            package_path: Path of the downloaded package file for the provided component.

        Raises:
            HopprPluginError: If an existing hash doesn't match.
        """
        generated_hashes = self._get_artifact_hashes(package_path)
        try:
            comp.update_hashes(generated_hashes)
        except ValueError as ex:
            raise HopprPluginError(f"Hash for {comp.name} does not match expected hash.") from ex

    def _validate_hashes(self, comp: Component, metadata_url: str, creds: CredentialRequiredService | None):
        """Validates the component hashes against the hashes in the package metadata provided by the Cargo server.

        Args:
            comp: Component object whose hashes will be validated.
            metadata_url: URL to the JSON document containing the package metadata.
            creds: Credentials used for HTTP requests.

        Raises:
            HopprPluginError: If hash validation fails.
        """
        try:
            validation_hash = self._get_package_checksum(metadata_url, creds)
        except (HopprPluginError, ValueError) as ex:
            self.get_logger().error(msg=str(ex), indent_level=2)
            raise HopprPluginError() from ex

        if not comp.contains_hashes([validation_hash]):
            raise HopprPluginError(f"Hash validation failed for {comp.name}.")

    @override
    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        purl = hoppr.utils.get_package_url(comp.purl)

        package_metadata_url, package_download_url = self._get_download_urls(purl, repo_url)

        self.get_logger().info(msg=f"Copying Cargo package from {package_download_url}", indent_level=2)

        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        package_out_path = target_dir / f"{purl.name}-{purl.version}.crate"

        try:
            self._download_package(comp.purl, package_download_url, package_out_path, creds)
            self._update_hashes(comp, package_out_path)
            self._validate_hashes(comp, package_metadata_url, creds)
        except HopprPluginError as ex:
            return Result.fail(message=str(ex))

        self.set_collection_params(comp, repo_url, package_out_path)
        return Result.success(return_obj=comp)



"""Collector plugin to copy artifacts using the Nexus API."""

from __future__ import annotations

import re
import time

from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import urlparse

import jmespath
import requests

from pydantic import SecretStr
from requests.auth import HTTPBasicAuth

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models.types import PurlType, RepositoryUrl
from hoppr.net import download_file
from hoppr.result import Result

if TYPE_CHECKING:
    from packageurl import PackageURL

    from hoppr.models import HopprContext
    from hoppr.models.credentials import CredentialRequiredService
    from hoppr.models.sbom import Component


class CollectNexusSearch(
    SerialCollectorPlugin,
    # Unless specified by the user, all types except `git`, `github`, and `gitlab` are supported
    supported_purl_types=[value for value in PurlType.values() if value not in {"git", "github", "gitlab"}],
):
    """Class to copy artifacts using the Nexus API."""

    def get_version(self) -> str:  # noqa: D102
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.supported_purl_types = (self.config or {}).get("purl_types", type(self).supported_purl_types)

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """Collect artifact from Nexus."""
        auth: HTTPBasicAuth | None = None
        if creds is not None and isinstance(creds.password, SecretStr):
            auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())

        nexus_url = self._parse_nexus_url(repo_url)[0]

        if not CollectNexusSearch.is_nexus_instance(nexus_url, auth):
            return Result.fail(f"{nexus_url} is not a Nexus instance")

        purl = hoppr.utils.get_package_url(comp.purl)

        try:
            source_urls = self.get_download_urls(purl, repo_url, auth)
        except ValueError as ex:
            self.get_logger().error(msg=str(ex))
            return Result.fail(message=str(ex))

        if not source_urls:
            msg = f"No artifacts found in Nexus instance {repo_url} for purl {comp.purl}"
            self.get_logger().error(msg, indent_level=2)
            return Result.fail(msg)

        target_dir = self._directory_for_nexus(purl, source_urls[0])

        for source_url in source_urls:
            self.get_logger().info(
                msg=f"Collecting from {source_url}",
                indent_level=2,
            )

            file_name = Path(urlparse(source_url).path).parts[-1]

            response = download_file(source_url, str(target_dir / file_name))
            nexus_result = Result.from_http_response(response)

            self.get_logger().info("Download Result: %s", nexus_result, indent_level=3)

            if not nexus_result.is_success():
                return nexus_result

        self.set_collection_params(comp, repo_url, target_dir / file_name)
        return Result.success(return_obj=comp)

    @staticmethod
    def _parse_nexus_url(repo_url: str) -> tuple[str, str | None]:
        nexus_url = repo_url
        nexus_repo = None
        if repo_specified := re.search(r"(https?://.*?)/repository/([^/]*)/?.*$", nexus_url):
            nexus_url = repo_specified[1]
            nexus_repo = repo_specified[2]

        return nexus_url, nexus_repo

    def _directory_for_nexus(self, purl: PackageURL, url: str) -> Path:
        nexus_repo = ""
        path = ""
        if repo_match := re.search(r"(.*?/repository/.*?)(/.*)?/(.*)", url):
            nexus_repo = repo_match[1]
            path = repo_match[2]
            if path is not None:
                path = path[1:]

        subdir = None
        match purl.type:
            case "docker" | "generic" | "maven" | "raw" | "rpm":
                subdir = path
            case "helm" | "pypi":
                subdir = f"{purl.name}_{purl.version}"

        return self.directory_for(purl.type, nexus_repo, subdir=subdir)

    @staticmethod
    def is_nexus_instance(repo_url: str, auth: HTTPBasicAuth | None = None) -> bool:
        """Checks whether or not the repo_url refers to a Nexus instance."""
        test_url = RepositoryUrl(url=repo_url) / "service" / "rest" / "v1" / "status"

        for attempt in range(3):
            if attempt > 0:
                time.sleep(5)

            response = requests.get(
                f"{test_url}",
                auth=auth,
                allow_redirects=True,
                stream=True,
                verify=True,
                timeout=60,
            )

            if response.status_code < 300:
                return True
            if response.status_code < 500:
                return False

        return False

    @staticmethod
    def _get_search_params(purl: PackageURL, nexus_repo: str | None) -> tuple[dict[str, str], list[dict[str, str]]]:
        """Returns search parameters for `get_download_urls`.

        Args:
            purl: Package URL object.
            nexus_repo: Nexus repository URL.

        Returns:
            Base parameters and additional search parameters for `get_download_urls`.
        """
        additional_search_params: list[dict[str, str]] = [{}]

        match purl.type:
            case "deb":
                nexus_format = "apt"
            case "gem":
                nexus_format = "rubygems"
            case "golang":
                nexus_format = "go"
            case "generic" | "raw":
                nexus_format = "raw"
            case "maven":
                nexus_format = "maven2"
                additional_search_params = [
                    {"maven.extension": "jar", "maven.classifier": ""},
                    {"maven.extension": "jar", "maven.classifier": "sources"},
                    {"maven.extension": "pom"},
                ]
            case "rpm" | "yum":
                nexus_format = "yum"
                arch = purl.qualifiers.get("arch")
                additional_search_params = [{"yum.architecture": arch}] if arch is not None else [{}]
            case _:
                nexus_format = purl.type

        base_params = {"sort": "version", "name": purl.name, "format": nexus_format}

        if purl.version is not None:
            base_params["version"] = purl.version

        if nexus_repo is not None:
            base_params["repository"] = nexus_repo

        return base_params, additional_search_params

    def get_download_urls(self, purl: PackageURL, repo_url: str, auth: HTTPBasicAuth | None = None) -> list[str]:
        """Retrieves all urls to be retrieved from Nexus for this component."""
        nexus_url, nexus_repo = CollectNexusSearch._parse_nexus_url(repo_url)
        search_url = RepositoryUrl(url=nexus_url) / "service" / "rest" / "v1" / "search" / "assets"

        base_params, additional_search_params = self._get_search_params(purl, nexus_repo)

        url_list: list[str] = []

        for extra_search_params in additional_search_params:
            response = requests.get(
                f"{search_url}",
                auth=auth,
                allow_redirects=True,
                stream=True,
                verify=True,
                timeout=60,
                params=base_params | extra_search_params,
            )

            try:
                response.raise_for_status()
            except requests.HTTPError:
                continue

            download_urls: list[str] = jmespath.search(
                expression=f"items[].downloadUrl | [? starts_with(@, '{repo_url}')]",
                data=response.json(),
            )

            match download_urls:
                case [url]:
                    url_list.append(url)
                case [url, *extra] if extra:
                    disambiguated_download_urls = _disambiguate_urls(repo_url, download_urls)
                    download_urls_bullets = "\n".join(f"  \u2022\u00a0{url}" for url in disambiguated_download_urls)

                    raise ValueError(
                        f'More than one result found for "{purl.name}" when searching "{repo_url}". '
                        "Please include a more specific repository URL in your manifest file. "
                        f"Either one of these should work:\n{download_urls_bullets}\n"
                    )
                case []:
                    self.get_logger().warning(
                        'No results found for "%s" when searching "%s".', purl.name, repo_url, indent_level=2
                    )

                    self.get_logger().debug("response:", indent_level=2)
                    self.get_logger().debug("url: %s", response.url, indent_level=3)
                    self.get_logger().debug("status code: %s", response.status_code, indent_level=3)
                    self.get_logger().debug("content: %s", response.content, indent_level=3)

        return url_list

    @classmethod
    def get_attestation_products(cls, config: dict | None = None) -> list[str]:  # noqa: D102
        products: list[str] = []

        if config is not None and "purl_types" in config:
            products.extend(f"{purl_type}/*" for purl_type in config["purl_types"])
        else:
            products.extend(
                f"{purl_type}/*" for purl_type in PurlType if str(purl_type) not in ["git", "github", "gitlab"]
            )

        return products


def _disambiguate_urls(base_url: str, urls: list[str]) -> list[str]:
    """Trims the URLs in the provided list after their first unambiguous subdirectory.

    Args:
        base_url: Known commonly-shared starting substring between the provided URLs.
        urls: URLs to disambiguate. Each URL is assumed to start with `base_url`, have one or
            more subdirectories and end with a filename.

    Raises:
        ValueError: If the URLs cannot be disambiguated.

    Returns:
        The provided URLs until their first unambiguous subdirectory.
    """
    ambiguous_paths: list[list[str]] = []
    max_length = 0
    disambiguated_urls: list[str] = []
    base_url = base_url.rstrip("/")

    for url in urls:
        _, *subdirectories, _ = url.removeprefix(base_url).split("/")
        max_length = max(max_length, len(subdirectories))
        ambiguous_paths.append(subdirectories)

    for level in range(max_length):
        not_repeated: dict[str, list[str]] = {}
        for path in ambiguous_paths:
            if len(path) <= level:
                raise ValueError("Repository URLs could not be disambiguated.")

            not_repeated[path[level]] = path if path[level] not in not_repeated else []

        for path in filter(None, not_repeated.values()):
            disambiguated_urls.append("/".join([base_url, *path[: level + 1]]))
            ambiguous_paths.remove(path)

    if ambiguous_paths:
        raise ValueError("Repository URLs could not be disambiguated.")

    return disambiguated_urls



"""`generate` subcommand for `hopctl`."""

from __future__ import annotations

import os

from pathlib import Path
from runpy import run_module

from rich.console import Console
from rich.markdown import Markdown
from typer import Option, Typer

from hoppr import main

app = Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Generate [bold cyan]in-toto[/] keys/layout or schemas for Hoppr input files",
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="rich",
)


@app.command()
def layout(
    transfer_file: Path = Option(
        "transfer.yml",
        "-t",
        "--transfer",
        envvar="HOPPR_TRANSFER_CONFIG",
        help="Specify transfer config",
    ),
    project_owner_key_path: Path = Option(
        ...,
        "-pk",
        "--project-owner-key",
        envvar="HOPPR_PROJECT_OWNER_KEY",
        help="Path to key used to sign in-toto layout",
    ),
    functionary_key_path: Path = Option(
        ...,
        "-fk",
        "--functionary-key",
        envvar="HOPPR_FUNCTIONARY_KEY",
        help="Path to key used to sign in-toto layout",
    ),
    project_owner_key_prompt: bool = Option(
        False,
        "-p",
        "--prompt",
        rich_help_panel="Deprecated",
        envvar="HOPPR_PROJECT_OWNER_KEY_PROMPT",
        help="Prompt user for project owner key's password (use --project-owner-key-password instead)",
        show_default=False,
    ),
    project_owner_key_password: str = Option(
        None,
        "-pk-pass",
        "--project-owner-key-password",
        confirmation_prompt=True,
        envvar="HOPPR_PROJECT_OWNER_KEY_PASSWORD",
        help="Password for project owner key",
        hide_input=True,
        prompt_required=False,
        prompt="Enter project owner key password",
        show_default=False,
    ),
):  # pragma: no cover
    """Create in-toto layout based on transfer file."""
    main.generate_layout(**locals())


@app.command(hidden=True)
def schemas(
    output_dir: Path = Option(
        Path("schema"),
        "-o",
        "--output-dir",
        exists=True,
        file_okay=False,
        help="Output directory for schema files [dim]\\[default: ./schema]",
        is_eager=True,
        resolve_path=True,
        show_default=False,
    ),
):  # pragma: no cover
    """Generate JSON/YAML schemas for Hoppr manifest, credential, and transfer files."""
    os.chdir(output_dir)
    run_module(mod_name="hoppr.models", run_name="__main__")


@app.command(hidden=True)
def validate_config(
    output: Path = Option(
        Path("hoppr") / "models" / "validation" / "profiles" / "default.config.yml",
        "-o",
        "--output",
        dir_okay=False,
        help="Path to default validation config file to write",
        is_eager=True,
    ),
):  # pragma: no cover
    """Generate SBOM validation configuration file populated with default values."""
    from hoppr.models.validation.checks import ValidateConfig

    default_config = ValidateConfig().yaml()

    console = Console()
    console.print(f"Writing default SBOM validation configuration to {output}...")

    output.write_text(default_config, encoding="utf-8")

    markdown = Markdown(
        markup="\n".join([
            "```yaml",
            output.read_text(encoding="utf-8").strip("\n"),
            "```",
        ]),
        code_theme="github-dark",
    )

    console.print(markdown)



"""Collector plugin for helm charts."""

from __future__ import annotations

import re

from pathlib import Path
from typing import TYPE_CHECKING

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx

from pydantic import SecretStr

import hoppr.net
import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.exceptions import HopprLoadDataError, HopprPluginError
from hoppr.models.sbom import Component, Hash
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from hoppr.models import HopprContext
    from hoppr.models.credentials import CredentialRequiredService


class CollectHelmPlugin(
    SerialCollectorPlugin,
    supported_purl_types=["helm"],
    required_commands=["helm"],
    products=["helm/*"],
    system_repositories=[],
):
    """Class to copy helm charts."""

    def get_version(self) -> str:  # noqa: D102
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.required_commands = (self.config or {}).get("helm_command", self.required_commands)
        self.base_command = [self.required_commands[0], "pull"]

        system_repos_file = Path.home() / ".config" / "helm" / "repositories.yaml"
        if not self.context.strict_repos and system_repos_file.exists():
            system_repos: list[dict[str, str]] = []

            try:
                system_repos_dict = hoppr.utils.load_file(input_file_path=system_repos_file)
                if not isinstance(system_repos_dict, dict):
                    raise HopprLoadDataError("Incorrect format.")

                system_repos = system_repos_dict["repositories"]
            except HopprLoadDataError as ex:
                self.get_logger().warning(msg=f"Unable to parse Helm repositories file ({system_repos_file}): '{ex}'")

            self.system_repositories.extend(repo["url"] for repo in system_repos)

    def _get_provenance_sha(self, prov_file: Path) -> str | Result | None:
        """Open an optional provenance file of a helm package and extract the sha signature.

        Args:
            prov_file: Path reference to the provenance file

        Returns:
            str: signature hash given by the provenance file
        """
        if provenance_sha := re.search(r"sha256:([a-fA-F0-9]{64})", prov_file.read_text(encoding="utf-8")):
            return provenance_sha[1]

        raise HopprPluginError("Hash cannot be found in provenance file")

    def _verify_provenance_sha(self, comp: Component, prov_sha: str) -> None:
        """Given a provenance signature, see if it is in the component's hashes.

        Args:
            comp: artifact that will be compared against the provenance signature
            prov_sha: provenance signature given as a str

        Raises:
            HopprPluginError: if component hash doesn't match provenance hash
        """
        if Hash(alg=cdx.HashAlg.SHA_256, content=prov_sha) not in comp.hashes:
            raise HopprPluginError(f"Hash for {comp.name} does not match hash from provenance.")

    def _update_hashes(self, comp: Component, package_path: Path) -> None:
        """Update the hashes of a component with hashes generated from its package file.

        Args:
            comp: Component object whose hashes will be updated.
            package_path: Path of the downloaded package file for the provided component.

        Raises:
            HopprPluginError: If an existing hash doesn't match.
        """
        generated_hashes = self._get_artifact_hashes(package_path)
        try:
            comp.update_hashes(generated_hashes)
        except ValueError as ex:
            raise HopprPluginError(f"Hash for {comp.name} does not match expected hash.") from ex

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """Collect helm chart."""
        purl = hoppr.utils.get_package_url(comp.purl)

        target_dir = self.directory_for(purl.type, repo_url, subdir=f"{purl.name}_{purl.version}")
        package_out_path = target_dir / f"{purl.name}-{purl.version}.tgz"

        run_result = None
        for subdir in ["", purl.name]:
            source_url = RepositoryUrl(url=repo_url) / subdir

            self.get_logger().info(msg="Fetching helm chart:", indent_level=2)
            self.get_logger().info(msg=f"source: {source_url}", indent_level=3)
            self.get_logger().info(msg=f"destination: {target_dir}", indent_level=3)

            command = [
                *self.base_command,
                "--repo",
                f"{source_url}",
                "--destination",
                f"{target_dir}",
                purl.name,
                "--version",
                purl.version,
                *(["--debug"] if self.get_logger().is_verbose() else []),
            ]

            password_list = []

            if creds is not None and isinstance(creds.password, SecretStr):
                command = [
                    *command,
                    "--username",
                    creds.username,
                    "--password",
                    creds.password.get_secret_value(),
                ]

                password_list = [creds.password.get_secret_value()]

            run_result = self.run_command(command, password_list)

            if run_result.returncode == 0:
                self.get_logger().info("Complete helm chart artifact copy for %s version %s", purl.name, purl.version)

                # Check for provenance
                command = [*command, "--prov"]
                self.run_command(command, password_list)
                file_path = target_dir / f"{purl.name}-{purl.version}.tgz"

                try:
                    self._update_hashes(comp, file_path)
                    self._verify_provenance_sha(
                        comp,
                        self._get_provenance_sha(file_path.with_suffix(".tgz.prov")),  # type: ignore[arg-type]
                    )
                except FileNotFoundError:
                    self.get_logger().info("Target artifact has no provenance file")
                except HopprPluginError as ex:
                    return Result.fail(str(ex))

                self.get_logger().info(msg="Fetching helm chart:", indent_level=2)
                self.set_collection_params(comp, repo_url, package_out_path)

                return Result.success(return_obj=comp)

        msg = f"Failed to download {purl.name} version {purl.version} helm chart from {repo_url}."

        if run_result is not None and "404 Not Found" in str(run_result.stderr):
            self.get_logger().debug(msg=msg, indent_level=2)
            return Result.fail(f"{msg} Chart not found.")

        self.get_logger().debug(msg=msg, indent_level=2)
        return Result.retry(msg)



"""Base model for `hopctl validate sbom` configuration models."""

from __future__ import annotations

import importlib
import inspect
import logging

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, ClassVar, TypeAlias, TypeGuard

import hoppr_cyclonedx_models.base as cdx_base

from pydantic import ConstrainedStr, Field, create_model, root_validator
from typing_extensions import Self

from hoppr.models.base import CycloneDXBaseModel, HopprBaseModel

if TYPE_CHECKING:
    from collections.abc import ValuesView

    from pydantic.fields import ModelField

    DictStrAny: TypeAlias = dict[str, Any]


def _is_list_model_field(field: ModelField) -> bool:
    return any(
        getattr(sub_field, "type_", None)
        and inspect.isclass(object=sub_field.type_)
        and issubclass(sub_field.type_, cdx_base.CycloneDXBaseModel)
        for sub_field in field.sub_fields or []
    )


def _is_model(obj: object) -> TypeGuard[type[cdx_base.CycloneDXBaseModel]]:
    return (
        inspect.isclass(obj)
        and obj.__module__.startswith(("hoppr.models", "hoppr_cyclonedx_models"))
        and issubclass(obj, cdx_base.CycloneDXBaseModel)
    )


class BaseCheckConfig(HopprBaseModel, alias_generator=lambda field_name: field_name.replace("_", "-"), frozen=True):
    """Base class for validation config definitions (singleton, immutable)."""

    __instance__: ClassVar[Self]

    def __new__(cls, *args, **kwargs) -> Self:  # noqa: D102
        if not getattr(cls, "__instance__", None):
            cls.__instance__ = super().__new__(cls)

        return cls.__instance__


BaseCheckConfig.update_forward_refs()


class BaseExcludeConfig(BaseCheckConfig):
    """Generates field definitions for items to exclude from validation.

    This class defines its fields dynamically based on the CycloneDX spec,
    so the final schema for the `exclude:` field is not generated until
    subclassed (i.e. until `__init_subclass__` is executed).
    """

    def __init_subclass__(cls) -> None:
        # Get all base `hoppr_cyclonedx_models.cyclonedx_1_5` model classes, excluding enum types and annotated types.
        # Classes in `hoppr_cyclonedx_models` are overridden by the redefined models in `hoppr` having the same name
        models: ValuesView[type[cdx_base.CycloneDXBaseModel]] = {
            model_name: model_cls
            for module_name in [
                "hoppr_cyclonedx_models.cyclonedx_1_5",
                "hoppr.models.affect",
                "hoppr.models.licenses",
                "hoppr.models.sbom",
            ]
            for model_name, model_cls in inspect.getmembers(
                object=importlib.import_module(name=module_name),
                predicate=_is_model,
            )
        }.values()

        field_definitions: dict[str, tuple[Any, Any]] = {
            field.name: (
                list[dict | ExcludePattern] | None,
                Field(
                    default=None,
                    title=field.field_info.title or field.name,
                    description=field.field_info.description or field.name,
                ),
            )
            for model in models
            for field in filter(_is_list_model_field, model.__fields__.values())
        }  # fmt: skip

        subclass_model = create_model(
            cls.__name__,
            __base__=BaseCheckConfig,
            __module__=__name__,
            **field_definitions,
        )  # type: ignore[call-overload]

        cls.__fields__ = subclass_model.__fields__

        return super().__init_subclass__()


BaseExcludeConfig.update_forward_refs()


class BaseValidator(CycloneDXBaseModel):
    """Base CycloneDX validator class."""

    # Mapping of check IDs to list of callback functions to report result
    observers: ClassVar[dict[object, Callable]] = {}

    @root_validator(pre=True, allow_reuse=True)
    @classmethod
    def remove_if_empty(cls, values: DictStrAny) -> DictStrAny:
        """Remove keys with falsy values to ensure they are counted as missing during validation."""
        for key, value in dict(values).items():
            if not value:
                values.pop(key, None)

        return values

    @classmethod
    def notify(cls, msg: str, log_level: int = logging.INFO, indent_level: int = 0) -> None:
        """Call the callback function for all registered subscribers."""
        for callback in cls.observers.values():
            callback(log_level, msg, indent_level=indent_level)

    @classmethod
    def subscribe(cls, observer: object, callback: Callable) -> None:
        """Register an observer."""
        cls.observers[observer] = callback

    @classmethod
    def unsubscribe(cls, observer: object) -> None:
        """Unregister an observer."""
        cls.observers.pop(observer, None)


BaseValidator.update_forward_refs()


class ExcludePattern(ConstrainedStr):
    """String with regular expression match constraint."""

    min_length = 1
    regex = r"^regexp:\/.+\/[gmi]*|jmespath:.+|.+"
    strip_whitespace = True



"""Hoppr signer class."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, ClassVar

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec, ed25519, padding, rsa

from hoppr.exceptions import HopprPrivateKeyError

if TYPE_CHECKING:
    import abc

    from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes
    from cryptography.hazmat.primitives.serialization.ssh import SSHPrivateKeyTypes


class HopprSigner:
    """Class for managing hoppr digital signatures.

    The intent is to generate digitally signed sboms and bundles
    """

    SUPPORTED_PRIVATE_KEY_TYPES: ClassVar[list[abc.ABCMeta]] = [
        ed25519.Ed25519PrivateKey,
        rsa.RSAPrivateKey,
        ec.EllipticCurvePrivateKey,
    ]

    private_key: SSHPrivateKeyTypes | PrivateKeyTypes | None

    def __init__(
        self,
        sign: bool,
        functionary_key_path: Path | None = None,
        functionary_key_password: str | None = None,
    ) -> None:
        self.sign = sign

        if self.sign and functionary_key_path:
            password_bytes = functionary_key_password.encode(encoding="utf-8") if functionary_key_password else None

            key_text = functionary_key_path.read_text()
            if "BEGIN OPENSSH PRIVATE KEY" in key_text:
                self.private_key = serialization.load_ssh_private_key(
                    key_text.encode(),
                    password=password_bytes,
                    backend=default_backend(),
                )
            else:
                self.private_key = serialization.load_pem_private_key(
                    key_text.encode(),
                    password=password_bytes,
                    backend=default_backend(),
                )

            if not isinstance(self.private_key, tuple(self.SUPPORTED_PRIVATE_KEY_TYPES)):
                raise HopprPrivateKeyError(
                    "Unsupported Private Key type for signing, supported types are rsa, ed25519, ecdsa."
                )

    def sign_blobs(self, blob_files: list[Path]):
        """Wrapper method used for handling multiple blobs that leverages sign_blob."""
        if self.sign:
            for blob_file in blob_files:
                self.sign_blob(blob_file)

    def sign_blob(self, blob_file: Path):
        """Signer method used by hoppr processing."""
        if self.sign:
            # Load the contents of the file to be signed.
            payload = blob_file.read_bytes()

            signature = None

            match self.private_key:
                case rsa.RSAPrivateKey():
                    signature = self.private_key.sign(
                        payload,
                        padding.PKCS1v15(),
                        hashes.SHA256(),
                    )
                case ed25519.Ed25519PrivateKey():
                    signature = self.private_key.sign(payload)
                case ec.EllipticCurvePrivateKey():
                    signature = self.private_key.sign(payload, ec.ECDSA(hashes.SHA256()))

            if signature:
                signature_file = Path(f"{blob_file!s}.sig")
                signature_file.write_bytes(signature)



"""Utility functions for hoppr plug-ins."""

from __future__ import annotations

import urllib.parse

from shutil import which

from hoppr.result import Result


def dir_name_from_repo_url(repo: str) -> str:
    """Creates a directory name based on a repo url."""
    if not repo.endswith("://"):
        repo = repo.removesuffix("/")

    return urllib.parse.quote_plus(repo)


def repo_url_from_dir_name(dir_name: str) -> str:
    """Reconstitutes a repo url from a directory name generated by dir_name_from_repo_url."""
    return urllib.parse.unquote_plus(dir_name)


def get_missing_commands(cmd_list: list[str] | list[str | list[str]]) -> list[str]:
    """Determine which items from a list of required commands are unavailable."""
    flattened_cmd_list: list[str] = []

    for cmd in cmd_list:
        flattened_cmd_list.extend(cmd if isinstance(cmd, list) else [cmd])

    return [cmd for cmd in flattened_cmd_list if not which(cmd)]


def check_for_missing_commands(cmd_list: list[str] | list[str | list[str]]) -> Result:
    """Raise an exception if any of a list of required commands are unavailable."""
    return (
        Result.fail(
            f"The following required commands are unavailable: {', '.join(missing_cmds)}. Please install and try again."
        )
        if (missing_cmds := get_missing_commands(cmd_list))
        else Result.success()
    )



"""Collector plugin for maven artifacts."""

from __future__ import annotations

import importlib
import os

from pathlib import Path
from subprocess import CalledProcessError, CompletedProcess
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any

import jmespath
import xmltodict

from pydantic import SecretStr

import hoppr.net
import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.result import Result

if TYPE_CHECKING:
    from collections import OrderedDict

    from hoppr.models import HopprContext
    from hoppr.models.credentials import CredentialRequiredService
    from hoppr.models.sbom import Component


_MAVEN_DEP_PLUGIN = "org.apache.maven.plugins:maven-dependency-plugin:3.5.0"


class CollectMavenPlugin(
    SerialCollectorPlugin,
    required_commands=["mvn"],
    supported_purl_types=["maven"],
    products=["maven/*"],
    system_repositories=["https://repo.maven.apache.org/maven2"],
):
    """Collector plugin for maven artifacts."""

    def get_version(self) -> str:  # noqa: D102
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.extra_opts: list[str] = []

        if self.config is not None:
            self.required_commands = self.config.get("maven_command", self.required_commands)
            self.extra_opts = self.config.get("maven_opts", self.extra_opts)

        system_settings_file = Path("/") / "etc" / "maven" / "settings.xml"
        user_settings_file = Path.home() / ".m2" / "settings.xml"

        if not self.context.strict_repos:
            # Identify system repositories
            for settings_file in [system_settings_file, user_settings_file]:
                if settings_file.is_file():
                    settings_dict: OrderedDict[str, Any] = xmltodict.parse(
                        settings_file.read_text(encoding="utf-8"),
                        encoding="utf-8",
                        force_list={"profile", "repository"},
                    )

                    repo_urls: list[str] = jmespath.search(
                        expression="settings.profiles.profile[].repositories.repository[].url", data=settings_dict
                    )

                    for repo in repo_urls or []:
                        if repo not in self.system_repositories:
                            self.system_repositories.append(repo)

    def _get_maven_component(self, command: list[str], password_list: list[str], **kwargs) -> CompletedProcess[bytes]:
        full_command = command.copy()
        full_command.extend(f"-D{key}={value}" for key, value in kwargs.items())
        result = self.run_command(full_command, password_list)

        # The maven plugin does not recognize the 'destFileName' argument, so rename the file
        if result.returncode == 0:
            _, name, version, extension = kwargs.get("artifact", "::::").split(":")
            directory: Path = kwargs.get("outputDirectory", ".")
            old_name = f"{name}-{version}.{extension}"
            new_name = kwargs.get("destFileName", f"{name}_{version}.{extension}")

            (directory / old_name).rename(directory / new_name)

        return result

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """Copy a component to the local collection directory structure."""
        purl = hoppr.utils.get_package_url(comp.purl)
        artifact = f"{purl.namespace}:{purl.name}:{purl.version}"

        extension = purl.qualifiers.get("type", "tar.gz")
        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        package_out_path = target_dir / f"{purl.name}_{purl.version}.{extension}"

        self.get_logger().info(msg="Copying maven artifact:", indent_level=2)
        self.get_logger().info(msg=f"source: {repo_url}", indent_level=3)
        self.get_logger().info(msg=f"destination: {target_dir}", indent_level=3)

        settings_dict = {
            "settings": {
                "servers": {
                    "server": {
                        "id": "repoId",
                        "username": "${repo.login}",
                        "password": "${repo.pwd}",
                    }
                }
            }
        }

        with NamedTemporaryFile(mode="w+", encoding="utf-8") as settings_file:
            settings_file.write(xmltodict.unparse(input_dict=settings_dict, pretty=True))
            settings_file.flush()

            password_list = []

            defines = {
                "artifact": f"{artifact}:{extension}",
                "outputDirectory": target_dir,
                "destFileName": f"{purl.name}_{purl.version}.{extension}",
                "remoteRepositories": f"repoId::::{repo_url}",
            }

            if creds is not None and isinstance(creds.password, SecretStr):
                defines["repo.login"] = creds.username
                defines["repo.pwd"] = creds.password.get_secret_value()
                password_list = [creds.password.get_secret_value()]

            command = [
                self.required_commands[0],
                f"{_MAVEN_DEP_PLUGIN}:copy",
                f"--settings={settings_file.name}",
                *self.extra_opts,
                *(["--debug"] if self.get_logger().is_verbose() else []),
            ]

            run_result = self._get_maven_component(command, password_list, **defines)
            error_msg = f"Failed to download maven artifact {artifact} type={extension}"

            try:
                run_result.check_returncode()

                if extension != "pom":
                    defines["artifact"] = str(defines["artifact"]).replace(f":{extension}", ":pom")
                    defines["destFileName"] = str(defines["destFileName"]).replace(f".{extension}", ".pom")

                    error_msg = f"Failed to download pom for maven artifact {artifact}"
                    run_result = self._get_maven_component(command, password_list, **defines)
                    run_result.check_returncode()
            except CalledProcessError:
                self.get_logger().debug(msg=error_msg, indent_level=2)
                return Result.retry(message=error_msg)

        self.set_collection_params(comp, repo_url, package_out_path)

        return Result.success(return_obj=comp)


if os.getenv("HOPPR_EXPERIMENTAL"):
    module = importlib.import_module(name="hoppr.plugins.collect.maven")
    CollectMavenPlugin = module.CollectMavenPlugin  # type: ignore[misc]



"""`merge` subcommand for `hopctl`."""

from __future__ import annotations

import time

from pathlib import Path
from typing import TYPE_CHECKING

from rich.console import Group
from rich.live import Live
from rich.table import Table
from typer import BadParameter, Option, Typer

import hoppr.utils

from hoppr.cli.layout import HopprBasePanel, HopprSbomFilesLayout, console
from hoppr.cli.options import (
    basic_term_option,
    log_file_option,
    manifest_file_option,
    sbom_dirs_option,
    sbom_files_option,
    sbom_urls_option,
    verbose_option,
)
from hoppr.models.base import CycloneDXBaseModel
from hoppr.models.sbom import Sbom

if TYPE_CHECKING:
    from rich.console import Console, ConsoleOptions, RenderResult

app = Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Merge all properties of two or more SBOM files",
    invoke_without_command=True,
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="rich",
    subcommand_metavar="",
)


class HopprMergeSummary(HopprBasePanel):
    """`hopctl merge` results summary."""

    def __init__(self, *args, title: str | None = "Summary", **kwargs) -> None:
        self.total_components = 0
        self.merged_components = 0

        self.summary_table = Table(
            "File Name",
            "# Components",
            "Elapsed Time",
            box=None,
            pad_edge=False,
            show_edge=False,
        )

        self.before_after_table = Table.grid(padding=(0, 1))

        self.summary_group = Group(self.summary_table, self.before_after_table)

        super().__init__(self.summary_group, title, *args, **kwargs)

    def __rich_console__(self, console_: Console, options: ConsoleOptions) -> RenderResult:
        self.before_after_table.add_row()
        self.before_after_table.add_row("Total components before merge:", str(self.total_components))
        self.before_after_table.add_row("Total components after merge:", str(self.merged_components))

        if hoppr.utils.is_basic_terminal():
            console_.line()
            console_.rule(title="Summary", characters="=", style="[bold]rule.line")
            console_.print(self.summary_table)
            console_.print(self.before_after_table)
            return []

        return super().__rich_console__(console_, options)


def output_file_callback(output_file: Path | None) -> Path:
    """Auto-generate an output file name if not provided."""
    return output_file or Path.cwd() / f"hopctl-merge-{time.strftime('%Y%m%d-%H%M%S')}.json"


def _merge_sboms() -> Sbom:
    """Merge SBOMs into single object."""
    merged_sbom = Sbom()
    layout = HopprSbomFilesLayout()
    live_display = Live(layout, console=console, refresh_per_second=10)
    merged_sbom.subscribe(observer=layout, callback=layout.print)
    summary_panel = HopprMergeSummary()

    for sbom_ref, sbom in Sbom.loaded_sboms.items():
        summary_panel.total_components += len(sbom.components)
        layout.add_job(description=Path(str(sbom_ref)).name, sbom=sbom)

    if not hoppr.utils.is_basic_terminal():
        live_display.start(refresh=True)

    for task in layout.jobs_panel.progress_bar.tasks:
        layout.print(f"Merging {task.description}...")
        layout.start_job(task.description)

        merged_sbom.merge(task.fields["sbom"])

        layout.stop_job(task.description)
        layout.advance_job(task.description)

        summary_panel.summary_table.add_row(
            task.description,
            str(len(task.fields["sbom"].components)),
            f"{task.elapsed:.3f}",
        )

    layout.print("Writing merged SBOM...")

    layout.overall_progress.progress_bar.stop_task(layout.progress_task.id)

    summary_panel.merged_components = len(merged_sbom.components)

    if not hoppr.utils.is_basic_terminal():
        live_display.stop()

    console.print("\n", summary_panel)

    merged_sbom.unsubscribe(observer=layout)
    return merged_sbom


@app.callback()
def merge(
    manifest_file: Path = manifest_file_option,
    sbom_files: list[Path] = sbom_files_option,
    sbom_dirs: list[Path] = sbom_dirs_option,
    sbom_urls: list[str] = sbom_urls_option,
    output_file: Path = Option(
        None,
        "-o",
        "--output-file",
        callback=output_file_callback,
        dir_okay=False,
        exists=False,
        help="Path to output file [dim]\\[default: hopctl-merge-YYMMDD-HHMMSS.json]",
        resolve_path=True,
        show_default=False,
    ),
    deep_merge: bool = Option(
        False,
        "--deep-merge",
        help="Resolve and expand [cyan]externalReferences[/] in-place",
        show_default=False,
    ),
    flatten: bool = Option(
        False,
        "--flatten",
        help="Flatten nested [cyan]components[/] into single unified list",
        show_default=False,
    ),
    basic_term: bool = basic_term_option,
    log_file: Path = log_file_option,
    verbose: bool = verbose_option,
):
    """Merge SBOM files."""
    if not any([manifest_file, *sbom_dirs, *sbom_files, *sbom_urls]) or len(Sbom.loaded_sboms) < 2:
        raise BadParameter(
            "A minimum of two SBOM files must be provided via the "
            "--sbom, --sbom-dir, --sbom-url, and --manifest arguments."
        )

    CycloneDXBaseModel.deep_merge = deep_merge
    CycloneDXBaseModel.flatten = flatten

    merged_sbom = _merge_sboms()

    output_file.write_text(
        data=merged_sbom.json(exclude_none=True, exclude_unset=True, by_alias=True, indent=2),
        encoding="utf-8",
    )
